var Spectrogram = (function () {
'use strict';

function hsl_to_rgb(h, s, l) {
    function hue_to_rgb(p, q, t) {
        if (t < 0)   t += 1;
        if (t > 1)   t -= 1;
        if (t < 1/6) return p + (q - p) * 6 * t;
        if (t < 1/2) return q;
        if (t < 2/3) return p + (q - p) * (2/3 - t) * 6;
        return p;
    }

    var q = l < 0.5 ? l * (1 + s) : l + s - l * s;
    var p = 2 * l - q;
    return {
        r: hue_to_rgb(p, q, h + 1/3),
        g: hue_to_rgb(p, q, h),
        b: hue_to_rgb(p, q, h - 1/3)
    };
}

var colormap = (function () {
    var length = 65535;
    var ret = {
        length: length - 1,
        r: new Uint8ClampedArray(length),
        g: new Uint8ClampedArray(length),
        b: new Uint8ClampedArray(length)
    };

    function sigmoid(x, phi, tau) {
        return 1.0 / (1 + Math.exp(-(x - phi) / tau));
    }

    for (var i = 0; i < length; i++) {
        var x = i / length;

        // Hue goes from red at 0 degrees back to red at 360 degrees. For best
        // contrast, we want to go from red to blue, and skip magenta. Blue is
        // at 240/360 = 0.66
        var hue = (1 - x) * 0.66;

        // Fade to black towards zero
        var lum = 0.5 * sigmoid(x, 0.05, 0.02);
        lum *= 0.1 + 0.9 * x;

        var rgb = hsl_to_rgb(hue, 0.85, lum);

        ret.r[i] = 255 * rgb.r;
        ret.g[i] = 255 * rgb.g;
        ret.b[i] = 255 * rgb.b;
    }

    return ret;
})();

function draw_colormap(ctx) {
    for (var y = 0; y < ctx.height; y++) {
        var i = Math.round((1 - (y / ctx.height)) * colormap.length);

        ctx.fillStyle = 'rgb(' + colormap.r[i] +
                           ',' + colormap.g[i] +
                           ',' + colormap.b[i] +
                           ')';

        ctx.fillRect(0, y, ctx.width, 1);
    }
}

function make_spectrogram(height) {
    var max = Math.max, min = Math.min;
    var elem = document.createElement('canvas');
    var ctx = elem.getContext('2d');
    if (!ctx) {
        alert('This browser does not support the canvas element');
        return null;
    }

    var img;

    var methods = {
        draw: function draw_spectrogram(dest_ctx) {
            if (img) {
                ctx.putImageData(img, 0, 0);

                var sx = 0;
                var sy = img.height - dest_ctx.height;
                var sWidth = img.width;
                var sHeight = dest_ctx.height;
                var dx = 0;
                var dy = 0;
                var dWidth = dest_ctx.width;
                var dHeight = dest_ctx.height;
                dest_ctx.drawImage(elem, sx, sy, sWidth, sHeight,
                                         dx, dy, dWidth, dHeight);
            }
        },
        shift: function trace_data_shift() {
            // Shift the image data up by one pixel
            if (!img || !img.data) return;
            img.data.set(img.data.subarray(4 * img.width), 0);

            var last_row = (height-1) * img.width * 4;
            for (var x = 0; x < img.width; x++) {
                var img_index = last_row + 4 * x;
                img.data[img_index+0] = 0;
                img.data[img_index+1] = 0;
                img.data[img_index+2] = 0;
                img.data[img_index+3] = 0;
            }
        },
        new_trace_data: function new_trace_data(new_points) {
            if (!new_points || !new_points.length) {
                return;
            }

            if (!img || img.width != new_points.length) {
                ctx.width = new_points.length;
                ctx.height = height;
                img = ctx.getImageData(0, 0, ctx.width, ctx.height);
                elem.width = new_points.length;
                elem.height = height;
                methods.clear();
            }

            var img_data = img.data;
            var last_row = (height-1) * img.width * 4;

            for (var x = 0; x < img.width; x++) {
                var z = max(0, min(1, new_points[x]));
                var r, g, b, a;
                if (isNaN(z)) {
                    // Draw invalid trace points as transparent
                    r = g = b = a = 0;
                }
                else {
                    var cm_index = Math.floor(z * colormap.length);
                    var cm_index = max(0, min(colormap.length, cm_index));
                    r = colormap.r[cm_index];
                    g = colormap.g[cm_index];
                    b = colormap.b[cm_index];
                    a = 255;
                }

                var img_index = last_row + 4 * x;
                img_data[img_index+0] = r;
                img_data[img_index+1] = g;
                img_data[img_index+2] = b;
                img_data[img_index+3] = a;
            }
        },
        clear: function spectrogram_clear() {
            if (!img) return;
            for (var i = 0, len = img.data.length; i < len; i+= 4) {
                img.data[i] = 0;
                img.data[i+1] = 0;
                img.data[i+2] = 0;
                img.data[i+3] = 0;
            }
        },
        draw_colormap: draw_colormap
    };

    return methods;
}

return {
    create: make_spectrogram,
    draw_colormap: draw_colormap,
};
})();
