function ScpiClient(client) {
    'use strict';
    // The cache is used during the corner case where we have recieved a
    // message from the server notifying us of a seting change, but we have
    // not yet attached a handler. It's a little clunky to sequence it
    // strictly (i.e. attach all handlers before joining the setting-value
    // room)
    var cache = {};
    var callbacks = {};

    var string_trim_regex = /(^\s+|\s+$)/g;
    var query_regex = /(\?)$/;
    var single_scpi_regex = /^\*?[a-z1-9:]+$/i;

    function normalize_single_cmd(cmd) {
        // Accepts a SCPI command like ASDF:QWER? and returns asdf:qwer. This
        // is used to account for different capitalizations. Does not account
        // for long/short form.
        cmd = cmd.toLowerCase();
        cmd = cmd.replace(string_trim_regex, '');
        // Remove the trailing question mark
        cmd = cmd.replace(query_regex, '');

        if (cmd.match(single_scpi_regex)) {
            return cmd;
        }
        else {
            return undefined;
        }
    }

    function alternatives(s) {
        // Given a SCPI command with alternatives, specified by the vertical
        // pipe |, returns a list of alternatives.
        // e.g. alternatives('asdf:qwer|zxcv') -> ['asdf:qwer', 'zxcv']
        // This is more than simply a split in order to handle empty strings
        // and to encapsulate the notion of alternatives.
        return (s || '').split('|').filter(function (s) {
            return (s || '').length > 0;
        });
    }

    function some_alternative(s) {
        // Given an SCPI command with possible alternatives, returns some
        // command. e.g. some_alternative('asdf|qwer') -> asdf
        // This is more than s.split('|')[0] because it reports errors.
        var alts = alternatives(s);
        if (alts.length > 0) {
            return alts[0];
        }
        else {
            console.error('Failed some_alternative("' + s + '")');
            return undefined;
        }
    }

    client.onconnect(function () {
        // Clear any stale setting values
        cache = {};

        client.join('setting-value', function (msg) {
            // The messages from setting-value are well specified, since they
            // are generated automatically by machine. There is always exactly
            // one SCPI command in the shortest possible form, with no
            // trailing question mark.
            alternatives(msg.command).map(function (cmd) {
                dispatch_one(cmd, msg.value);
            });
        });        

        client.join('scpi-log', function (msg) {
            // Unlike the messages from setting-value, messages from scpi-log
            // are completely unspecified because they can be entered by a
            // user. We can get useful information from the message if we
            // check that it's (1) exactly one command and (2) it's a query.
            alternatives(msg.command).map(function (cmd) {
                dispatch_one(cmd, msg.response);
            });
        });


        function dispatch_one(cmd, response) {
            var cmd = normalize_single_cmd(cmd);

            if (cmd) {
                (callbacks[cmd] || []).map(function (callback) {
                    callback(response);
                });

                cache[cmd] = response;
            }
        }
    });

    function on(orig_cmd, func) {
        return alternatives(orig_cmd).map(function (cmd) {
            return on_one_cmd(cmd, func);
        });
    }

    function on_one_cmd(orig_cmd, func) {
        // Remove the trailing question mark
        var cmd = normalize_single_cmd(orig_cmd);

        if (!cmd) {
            console.error('Cannot listen for SCPI "' + cmd + '"');
            return;
        }

        if (typeof func !== 'function') {
            console.error('Bad callback for "' + cmd + '"');
            return;
        }

        callbacks[cmd] = callbacks[cmd] || [];
        callbacks[cmd].push(func);
        if (typeof cache[cmd] !== 'undefined') {
            func(cache[cmd]);
        }
    }

    function __send(endpoint, cmd, callback) {
        if (typeof callback === 'function') {
            client.send(endpoint, cmd, function (msg) {
                callback(msg.response);
            });
        }
        else {
            client.send(endpoint, cmd);
        }
    }

    function send(cmd, callback) {
        var alt = some_alternative(cmd)
        if (alt) {
            __send('scpi', alt, callback);
        }
    }

    function send_quiet(cmd, callback) {
        __send('scpi-quiet', cmd, callback);
    }

    function set_value(cmd, value) {
        var alt = some_alternative(cmd);
        if (alt) {
            __send('scpi', alt + ' ' + String(value));
        }
    }

    function query_and_listen(orig_cmd, func) {
        var cmd = normalize_single_cmd(orig_cmd);
        if (cmd) {
            on(cmd, func);
            send_quiet(cmd + '?');
        }
        else {
            console.error('Cannot query SCPI "' + orig_cmd + '"');
        }
    }

    function get(cmd) {
        return cache[cmd];
    }

    return {
        on: on,
        set_value: set_value,
        onconnect: client.onconnect,
        send: send,
        send_quiet: send_quiet,
        query_and_listen: query_and_listen,
        alternatives: alternatives,
        some_alternative: some_alternative,
        normalize: normalize_single_cmd,
        get: get,
    };
}

