'use strict'

/*
 * This jQuery plugin provides a way to show the user that the input they have
 * modified has not taken effect yet.
 *
 * The browser model does not use dependency injection, yet it does emit
 * objects asynchronously. Therefore, when an event is emitted, it is some
 * fixed event. We cannot have the browser emit a custom object. Similarly,
 * jQuery does not use dependency injection so when we call $('selector')
 * we always get back a plain jQuery object, not our custom object. Thus, we
 * have to use a non-obvious method dispatch pattern.
 *
 * Assuming you have selected some DOM object like so:
 *    var $elem = $('your selector here');
 *
 * then initialize the setting input like this:
 *    $elem.setting_input('init', init_type);
 * where init_type is 'number', 'frequency', or 'text'.
 *
 * To set the value from code, call the value method like so:
 *    $elem.setting_input('value', 1234);
 *
 * To get the textual value, call:
 *    $elem.setting_input('value');
 *
 * To listen for when the user commits (this is a complex interaction, moreso
 * for frequency inputs):
 *
 *    $elem.on('commit', function () { ... });
 *
 * Do not use the change event, as it will be emitted during many invalid or
 * intermediate states. It is not feasable to override the change event to
 * have the behavior of the commit event.
 *
 */
;(function ($) {
// Variables prepended with a dollar sign $ are jQuery objects

function is_escape_event(e) {
    return (e.which || e.keyCode || e.charCode) === 27;
}

function is_enter_event(e) {
    var k = e.which || e.keyCode || e.charCode;
    return (k === 10 || k === 13);
}

function keystr(e) {
    var key = e.which || e.keyCode || e.charCode;
    // String.fromCharCode expects a unicode number, for the keypad, keys 96-105 these are letters so they
    // need to be shifted back to numbers
    return e.key || String.fromCharCode((96 <= key && key <= 105)? key-48 : key);
}

function on_escape(elem, cb) {
    $(elem).on('keydown', function (e) {
        if (is_escape_event(e)) {
            e.preventDefault();
            cb();
        }
    });
}

if (typeof $.propHooks === 'object') {
    $.propHooks.disabled = {
        set: function (el, value) {
            if (el.disabled !== value) {
                el.disabled = value;
                $(el).trigger('disabled', value);
            }
        }
    };
}

function init_setting_input($input, init_args) {
    // This wrapper groups $input with the buttons, and in the frequency case
    // the frequency dropdown.
    var $wrapper = $('<span class="setting-input-wrapper"></span>');
    // This wrapper keeps the buttons together
    var $buttons = $('<span class="setting-input-buttons"></span>');
    $buttons.hide();

    var $commit = $('<button title="Commit">&#10003;</button>');
    var $cancel = $('<button title="Cancel">&#10007;</button>');

    $wrapper.insertBefore($input);

    $wrapper.append($input.detach());
    $wrapper.append($buttons);

    $buttons.append($commit);
    $buttons.append($cancel);

    var committed_value = '';
    $input.val(committed_value);

    init_args = init_args || {};
    if (init_args.init_wrapper) {
        init_args.init_wrapper($wrapper);
    }

    var methods = {
        get_ui_value: function get_ui_value() { return $input.val(); },
        set_ui_value: function set_ui_value(value) { $input.val(value); },
        handle_keydown: function handle_keydown () {},
        validate: function return_true() { return true; },
        value: function (v) {
            if (typeof v !== 'undefined') {
                methods.cancel();
                methods.set_ui_value(v);
                // UI values must always be strings
                committed_value = String(v);
            }
            return committed_value;
        }
    };

    methods.commit = function commit() {
        var this_value = methods.get_ui_value();

        if (methods.validate(this_value)) {
            if (this_value !== committed_value) {
                committed_value = this_value;
                methods.set_ui_value(this_value);
                $input.trigger('commit', [this_value]);
            }
            methods.set_dirty(false);
        }
        else {
            console.log('invalid');
        }
    }

    methods.set_dirty = function set_dirty(show) {
        $buttons[show ? 'show' : 'hide']();
    }

    methods.cancel = function cancel() {
        $input.blur();
        methods.set_ui_value(committed_value)
        $buttons.hide();
    }

    $input.on('keydown', function (e) {
        if (is_enter_event(e)) {
            methods.commit();
            e.preventDefault();
        }
        else {
            methods.handle_keydown(e);
        }
    });

    $commit.on('click', function () {
        methods.commit();
    });

    on_escape($input, methods.cancel);
    on_escape($commit, methods.cancel);
    on_escape($cancel, methods.cancel);
    $cancel.on('click', methods.cancel);

    $input.on('input', function() {
        var ui_value = methods.get_ui_value();

        // Browsers do not provide the value for an invalid numeric input.
        // Instead, they provide the empty string.  So it's insufficient to
        // simply check the value. This asks the browser if there does exist
        // some invalid input.
        var has_invalid_input = this.validity && !this.validity.valid;
        var dirty = (has_invalid_input || ui_value !== committed_value);
        methods.set_dirty(dirty);
        $commit.prop('disabled', !methods.validate(ui_value));
    });

    return methods;
}

function init_setting_input_number($input) {
    var methods = init_setting_input($input);
    $input.attr('type', 'number');
    $input.attr('step', 'any');

    methods.validate = function validate_float(val) {
        if (val === '') val = 'nan'; // Number('') === 0
        return !isNaN(Number(val));
    }

    return methods;
}

function init_setting_input_freq($input) {
    $input.addClass('freq-input');
    $input.attr('type', 'number');
    $input.attr('step', 'any');

    var units = {
        h: { label: 'Hz',  factor: 1e0 },
        k: { label: 'kHz', factor: 1e3 },
        m: { label: 'MHz', factor: 1e6 },
        g: { label: 'GHz', factor: 1e9 },
    }

    var $units = $('<select>');
    $units.addClass('freq-input-units');
    for (var k in units) {
        var option_elem = $('<option value="' + k + '">' +
                            units[k].label +
                            '</option>')
        $units.append(option_elem);
    }

    function valid_unit(u) {
        u = String(u)[0].toLowerCase();
        return typeof units[u] !== 'undefined';
    }

    function set_units(u) {
        u = String(u).toLowerCase();
        if (typeof units[u] !== 'undefined') {
            $units.val(u);
            return true;
        }
        else {
            return false;
        }
    }

    function get_units() {
        var ret = units[$units.val()];
        if (typeof ret !== 'undefined') {
            return ret;
        }
        else {
            throw 'Internal error, unknown unit "' + $units.val() + '"';
        }
    }

    var methods = init_setting_input($input, {
        init_wrapper: function init_wrapper($wrapper) {
            $units.insertAfter($input);
        }
    });
        
    methods.get_ui_value = function get_ui_value() {
        return $input.val() + ' ' + get_units().label;
    }

    methods.handle_keydown = function handle_keydown(e) {
        if (set_units(keystr(e))) {
            e.preventDefault();
            methods.commit();
        }
    }

    var commit_on_units_change = false;
    $input.on('input', function () {
        commit_on_units_change = true;
    });

    $units.on('keypress', function (e) {
        if (is_enter_event(e)) {
            methods.commit();
        }
    });

    $units.on('change', function () {
        if (commit_on_units_change) {
            methods.commit();
        }
        else {
            commit_on_units_change = true;
            if ($input.val() !== '') {
                methods.set_dirty(true);
            }
        }
    });

    on_escape($units, methods.cancel);

    function method_chain(obj, method, additionally) {
        var super_method = obj[method];
        obj[method] = function () {
            super_method();
            additionally();
        };
    }

    method_chain(methods, 'cancel', function frequency_cancel() {
        commit_on_units_change = false;
    });

    method_chain(methods, 'commit', function frequency_cancel() {
        commit_on_units_change = false;
    });

    $input.on('disabled', function (v) {
        $units.prop('disabled', v);
    });

    methods.set_ui_value = function set_ui_value(new_val) {
        var toks = String(new_val).split(/\s+/);
        if (toks.length === 0) {
            return;
        }
        else if (toks.length === 1) {
            // Assume unit Hz
            var val_hz = Number(new_val || 'nan');
            var sign = Math.sign(val_hz);
            val_hz *= sign;
            if (isNaN(val_hz)) {
                $input.val('');
                return;
            }
            else if (val_hz > 1e9) {
                set_units('g');
            }
            else if (val_hz > 1e6) {
                set_units('m');
            }
            else if (val_hz > 1e3) {
                set_units('k');
            }
            else {
                set_units('h');
            }

            $input.val(sign * val_hz / get_units().factor);
        }
        else if (toks.length === 2) {
            if (!isNaN(Number(toks[0]))) {
                if (set_units(toks[1])) {
                    $input.val(toks[0]);
                }
            }
            else {
                throw 'Cannot set frequency to ' + new_val;
            }
        }
        else {
            throw 'Cannot set frequency to ' + new_val;
        }
    }

    methods.validate = function validate_float(val) {
        var toks = String(val).split(/\s+/);
        if (toks.length === 1) {
            return !isNaN(Number(toks[0] || 'nan'));
        }
        else if (toks.length === 2) {
            return !isNaN(Number(toks[0] || 'nan')) && valid_unit(toks[1]);
        }
        else {
            return false;
        }
    }

    return methods;
}

function init_setting_input_memory($input) {
    $input.addClass('memory-input');
    $input.attr('type', 'number');
    $input.attr('step', 'any');

    var raw_units = $input.attr('units');

    var units = {};

    //Assume ordered
    var desired_units = raw_units.split(',');

    var possible_units = {
        b: { label: 'B', factor: 1 },
        kb: { label: 'KB', factor: 1e3 },
        mb: { label: 'MB', factor: 1e6 },
        gb: { label: 'GB', factor: 1e9 },
    };

    units[desired_units[0]] = possible_units[desired_units[0]];

    for (var i = 0; i < desired_units.length; i++) {
       units[desired_units[i]] = { 
           label: possible_units[desired_units[i]].label, 
           factor: possible_units[desired_units[i]].factor
       };
    }   

    var $units = $('<select>');
    $units.addClass('memory-input-units');
    for (var k in units) {
        var option_elem = $('<option value="' + k + '">' +
                            units[k].label +
                            '</option>')
        $units.append(option_elem);
    }

    function valid_unit(u) {
        u = String(u).toLowerCase();
        return typeof units[u] !== 'undefined';
    }

    function set_units(u) {
        u = String(u).toLowerCase();
        if (typeof units[u] !== 'undefined') {
            $units.val(u);
            return true;
        }
        else {
            return false;
        }
    }

    function get_units() {
        var ret = units[$units.val()];
        if (typeof ret !== 'undefined') {
            return ret;
        }
        else {
            throw 'Internal error, unknown unit "' + $units.val() + '"';
        }
    }

    var methods = init_setting_input($input, {
        init_wrapper: function init_wrapper($wrapper) {
            $units.insertAfter($input);
        }
    });

    methods.get_ui_value = function get_ui_value() {
        return $input.val()*get_units().factor;
    }

    methods.handle_keydown = function handle_keydown(e) {
        // The single keypress for "milliseconds" is ambiguous with "minutes"
        // so it's not very sensible to automatically commit
    }

    var commit_on_units_change = false;
    $input.on('input', function () {
        commit_on_units_change = true;
    });

    $units.on('keypress', function (e) {
        if (is_enter_event(e)) {
            methods.commit();
        }
    });

    $units.on('change', function () {
        if (commit_on_units_change) {
            methods.commit();
        }
        else {
            commit_on_units_change = true;
            if ($input.val() !== '') {
                methods.set_dirty(true);
            }
        }
    });

    on_escape($units, methods.cancel);

    function method_chain(obj, method, additionally) {
        var super_method = obj[method];
        obj[method] = function () {
            super_method();
            additionally();
        };
    }

    method_chain(methods, 'cancel', function time_cancel() {
        commit_on_units_change = false;
    });

    method_chain(methods, 'commit', function time_cancel() {
        commit_on_units_change = false;
    });

    $input.on('disabled', function (v) {
        $units.prop('disabled', v);
    });

    methods.set_ui_value = function set_ui_value(new_val) {
        var toks = String(new_val).split(/\s+/);
        if (toks.length === 0) {
            return;
        }
        else if (toks.length === 1) {
            // Time value always return in milliseconds from SCPI 
            var converted_val = Number(new_val || 'nan');
            var sign = Math.sign(converted_val);
            converted_val *= sign;
            if (isNaN(converted_val)) {
                $input.val('');
                return;
            }
            else if ((converted_val > possible_units['gb'].factor) && (units['gb'] !== "undefined")) {
                set_units('gb');
            }
            else if ((converted_val > possible_units['mb'].factor) && (units['mb'] !== "undefined")) {
                set_units('mb');
            }
            else if ((converted_val > possible_units['kb'].factor) && (units['kb'] !== "undefined")) {
                set_units('kb');
            }
            else
                set_units('b');

            $input.val(sign * converted_val / get_units().factor);
        }
        else if (toks.length === 2) {
            if (!isNaN(Number(toks[0]))) {
                if (set_units(toks[1])) {
                    $input.val(toks[0]);
                }
            }
            else {
                throw 'Cannot set to ' + new_val;
            }
        }
        else {
            throw 'Cannot set to ' + new_val;
        }
    }

    methods.validate = function validate_float(val) {
        var toks = String(val).split(/\s+/);
        if (toks.length === 1) {
            return !isNaN(Number(toks[0] || 'nan'));
        }
        else if (toks.length === 2) {
            return !isNaN(Number(toks[0] || 'nan')) && valid_unit(toks[1]);
        }
        else {
            return false;
        }
    }

    return methods;
 
}

function init_setting_input_time($input) {
    $input.addClass('time-input');
    $input.attr('type', 'number');
    $input.attr('step', 'any');
    var raw_units = $input.attr('units');

    var units = {};

    //Assume ordered.
    var desired_units = raw_units.split(',');

    var possible_units = {
        us:  { label: 'us',  factor: 1 },
        ms:  { label: 'ms',  factor: 1e3 },
        s:   { label: 's',   factor: 1e6 },
        min: { label: 'min', factor: 6e7 },
        hr:  { label: 'hr',  factor: 3.6e9 },
        day: { label: 'day', factor: 86.4e9 },
    };


    units[desired_units[0]] = possible_units[desired_units[0]];

    var divider = units[desired_units[0]].factor; 

    for (var i = 0; i < desired_units.length; i++) {
       units[desired_units[i]] = { 
           label: possible_units[desired_units[i]].label, 
           factor: possible_units[desired_units[i]].factor
       };
    }   

    

    var $units = $('<select>');
    $units.addClass('time-input-units');
    for (var k in units) {
        var option_elem = $('<option value="' + k + '">' +
                            units[k].label +
                            '</option>')
        $units.append(option_elem);
    }

    function valid_unit(u) {
        u = String(u).toLowerCase();
        return typeof units[u] !== 'undefined';
    }

    function set_units(u) {
        u = String(u).toLowerCase();
        if (typeof units[u] !== 'undefined') {
            $units.val(u);
            return true;
        }
        else {
            return false;
        }
    }

    function get_units() {
        var ret = units[$units.val()];
        if (typeof ret !== 'undefined') {
            return ret;
        }
        else {
            throw 'Internal error, unknown unit "' + $units.val() + '"';
        }
    }

    var methods = init_setting_input($input, {
        init_wrapper: function init_wrapper($wrapper) {
            $units.insertAfter($input);
        }
    });

    methods.get_ui_value = function get_ui_value() {
        return $input.val() + ' ' + get_units().label;
    }

    methods.handle_keydown = function handle_keydown(e) {
        // The single keypress for "milliseconds" is ambiguous with "minutes"
        // so it's not very sensible to automatically commit
    }

    var commit_on_units_change = false;
    $input.on('input', function () {
        commit_on_units_change = true;
    });

    $units.on('keypress', function (e) {
        if (is_enter_event(e)) {
            methods.commit();
        }
    });

    $units.on('change', function () {
        if (commit_on_units_change) {
            methods.commit();
        }
        else {
            commit_on_units_change = true;
            if ($input.val() !== '') {
                methods.set_dirty(true);
            }
        }
    });

    on_escape($units, methods.cancel);

    function method_chain(obj, method, additionally) {
        var super_method = obj[method];
        obj[method] = function () {
            super_method();
            additionally();
        };
    }

    method_chain(methods, 'cancel', function time_cancel() {
        commit_on_units_change = false;
    });

    method_chain(methods, 'commit', function time_cancel() {
        commit_on_units_change = false;
    });

    $input.on('disabled', function (v) {
        $units.prop('disabled', v);
    });

    methods.set_ui_value = function set_ui_value(new_val) {
        var toks = String(new_val).split(/\s+/);
        if (toks.length === 0) {
            return;
        }
        else if (toks.length === 1) {
            // Time value always return in milliseconds from SCPI 
            var time_val = Number(new_val || 'nan');
            var sign = Math.sign(time_val);
            time_val *= sign;
            if (isNaN(time_val)) {
                $input.val('');
                return;
            }
            else if ((time_val > possible_units['day'].factor) && (units['day'] !== "undefined")) {
                set_units('day');
            }
            else if ((time_val > possible_units['hr'].factor) && (units['hr'] !== "undefined")) {
                set_units('hr');
            }
            else if ((time_val > possible_units['min'].factor) && (units['min'] !== "undefined")) {
                set_units('min');
            }
            else if ((time_val > possible_units['s'].factor) && (units['s'] !== "undefined")) {
                set_units('s');
            }
            else if ((time_val > possible_units['ms'].factor) && (units['ms'] !== "undefined")) {
                set_units('ms');
            }
            else
                set_units('us');

            $input.val(sign * time_val / get_units().factor);
        }
        else if (toks.length === 2) {
            if (!isNaN(Number(toks[0]))) {
                if (set_units(toks[1])) {
                    $input.val(toks[0]);
                }
            }
            else {
                throw 'Cannot set time to ' + new_val;
            }
        }
        else {
            throw 'Cannot set time to ' + new_val;
        }
    }

    methods.validate = function validate_float(val) {
        var toks = String(val).split(/\s+/);
        if (toks.length === 1) {
            return !isNaN(Number(toks[0] || 'nan'));
        }
        else if (toks.length === 2) {
            return !isNaN(Number(toks[0] || 'nan')) && valid_unit(toks[1]);
        }
        else {
            return false;
        }
    }

    return methods;
}

$.fn.setting_input = function (method) {
    var methods = $(this).data('setting-input');

    if (method === 'init') {
        if (typeof methods === 'undefined') {
            var init_funcs = {
                text: init_setting_input,
                number: init_setting_input_number,
                frequency: init_setting_input_freq,
                time: init_setting_input_time,
                memory: init_setting_input_memory,
            };

            var typ = arguments[1] || 'text';
            var init_func = init_funcs[typ];

            if (typeof init_func === 'function') {
                methods = init_func($(this));
                if (typeof methods === 'undefined') {
                    throw 'init for ' + typ + ' did not return methods';
                }
                $(this).data('setting-input', methods);
            }
            else {
                $.error('Unknown method for jQuery.setting_input ' + typ);
            }
        }
        else {
            $.error('Element is already initialized');
        }
    }
    else {
        if (typeof methods === 'undefined') {
            throw 'Uninitialized jQuery.setting_input';
        }
        else if (typeof methods[method] === 'function') {
            var ret = methods[method].apply(this,
                Array.prototype.slice.call(arguments, 1));
            if (typeof ret !== 'undefined') {
                return ret;
            }
        }
        else {
            $.error('Unknown method for jQuery.setting_input ' + method);
        }
    }

    return $(this);
};

$.fn.setting_value = function () {
    var args = Array.prototype.slice.call(arguments);
    args.unshift('value');
    $.fn.setting_input.apply(this, args);
};

})(jQuery);
