'use strict';
function setup_y_scale_commands(scpi, trace_data, spectrum) {
    scpi.on('DISP:WIND:TRAC:Y:SCAL:RLEV', function (n) {
        trace_data.set_ref_level(parseFloat(n));
    });

    scpi.on('UNIT:POW', function (n) {
        trace_data.set_units(n);
        var limitLabel = document.getElementById('limit-amplitude-label');
        if (limitLabel) {
            limitLabel.innerHTML = 'Amplitude (' + pretty_format_amplitude_unit(n) + ')';
        }
    });

    scpi.on('disp:wind:trac:y:scal:pdiv', function (scale) {
        set_amplitude_span(scale, spectrum, trace_data);
    });
}

function set_amplitude_span(scale, spectrum, trace_data){
    var scale = parseFloat(scale);
    var ndivs = spectrum.num_divs();
    if (!isNaN(scale)) {
        trace_data.set_amplitude_span(scale * ndivs);
    }
}

function setup_bandwidth_ratio(scpi) {
    function aux(selector, cmd) {
        var $el = $(selector).setting_input('init', 'number');

        $el.on('commit', function (_, val) {
            scpi.set_value(cmd, 1 / parseFloat(val));
        });

        scpi.on(cmd, function (n) {
            var value = Math.round(1 / parseFloat(n));
            if (!isNaN(value)) {
                $el.setting_value(value);
            }
        });
    }

    aux('#SpanRBWRatio', 'SENS:BAND:RES:RAT');
    aux('#VBWRBWRatio', 'SENS:BAND:VID:RAT');
}

function setup_frequency_commands(scpi, trace_data) {
    scpi.on('SENS:FREQ:STAR', function (n) {
        trace_data.frequency.start = parseFloat(n);
    });

    scpi.on('SENS:FREQ:STOP', function (n) {
        trace_data.frequency.stop = parseFloat(n);
    });

    scpi.on('SENS:FREQ:CENT', function (n) {
        trace_data.frequency.center = parseFloat(n);
    });

    scpi.on('SENS:FREQ:SPAN', function (n) {
        trace_data.frequency.span = parseFloat(n);
    });
}

function trace_refresh_rate_message()
{
    var is_checked = false;

    $('#set-manual-refresh-rate').change(function(){
         is_checked = $('#set-manual-refresh-rate').is(':checked');
    });

    return {
        get: function(){
            var ret = [];

            if(Boolean(is_checked)){
                ret.push({ style: 'red', label: 'Data throttling is enabled' });
            }

            return ret;
        }
    }
}

function CalMessages(scpi) {
    var rlc = true;
    var ifgc = true;
    var antc = true;
    var shap = true;
    var tempcomp = true;

    var $antenna_mux = $('[data-scpi="deb:cal:antc:stat"]');
    var enabled_for = { on: true, off: true };

    scpi.on('DEB:CAL:RLC:STAT', function (s) {
        rlc = s === '1';
    });

    scpi.on('DEB:CAL:IFGC:STAT', function (s) {
        ifgc = s === '1';
    });

    scpi.on('DEB:CAL:ANTC:STAT', function (s) {
         antc = (s === 'ON' || s === 'NODE')? true:false;
         $antenna_mux.prop('disabled', !enabled_for[s.toLowerCase()]);
    });

    scpi.on('DEB:TEMP:COMP:STAT', function (s) {
        tempcomp = s === '1';
    });

    scpi.on('DEB:CAL:SHAP:STAT', function (s) {
        shap = s === '1';
    });
    return {
        get: function () {
            var ret = [];
            function __push(label) {
                ret.push({ style: 'red', label: label });
            }

            if (!rlc)  __push('RL Calibration is Off');
            if (!ifgc) __push('IF Calibration is Off');
            if (!antc) __push('Antenna Mux Calibration is Off');
            if (!shap) __push('Shape Calibration is Off');
            if (!tempcomp) __push('Temperature Compensation is Off');
            return ret;
        }
    }
}

function setup_overpower(scpi) {
    var modal = modal_overlay('#overpower');
    $('#close-relay-button').on('click', function () {
        scpi.send('inp:opow:rel:stat clos');
        modal.hide();
    });

    scpi.on('INP:OPOW:REL:STAT', function (n) {
        if (n === "OPEN") {
            modal.show();
        }
        else if (n === "CLOS") {
            modal.hide();
        }
    });
}

function get_all_json(paths, cb) {
    var pending = paths.length;
    var result = {};
    paths.map(function (path) {
        $.getJSON(path).done(function (data) {
            result[path] = data;
        }).always(function () {
            pending--;
            if (pending === 0) {
                cb(result);
            }
        });
    });
}

function bind_setting_change_handlers(scpi) {
    function init_select_text($el) {
        // Read out the <select> tags like a map and replace them with spans
        // so we can get rid of the confusing <select> styling
        var options = {};
        $el.children('option').each(function () {
            options[$(this).val()] = {
                text: $(this).text(),
                cls: $(this).attr('class')
            };
        });

        var scpi_cmd = $el.data('scpi');
        var span_elem = $("<span>");

        scpi.on(scpi_cmd, function (scpi_value) {
            var option = options[scpi_value] || {};
            span_elem.text(option.text);
            span_elem.addClass(option.cls);
            span_elem.addClass('scpi-text scpi-display');
        });

        $el.replaceWith(span_elem);
    }

    function init_checkbox($el) {
        var unchecked_value = String($el.data('scpi-unchecked')).toLowerCase();
        var checked_value = String($el.data('scpi-checked')).toLowerCase();

        $el.prop('indeterminate', true);
        var scpi_cmd = $el.data('scpi');

        scpi.on(scpi_cmd, function (scpi_value) {
            scpi_value = scpi_value.toLowerCase();
            scpi_value = scpi_value.replace('\n', '');

            var checked;
            if (scpi_value === unchecked_value) {
                checked = false;
            }
            else if (scpi_value === checked_value) {
                checked = true;
            }

            $el.prop('indeterminate', typeof checked === 'undefined');
            $el.prop('checked', checked);
        });

        //TODO once all the check-box settings are converted to boolean simplify the below condition
        $el.on('change', function () {
            if (!$el.prop('indeterminate')) {
                var val = $el.prop('checked') ?
                    (checked_value == 'true'?'1':checked_value) : (unchecked_value == 'false'?'0':unchecked_value);

                scpi.set_value(scpi_cmd, val);
            }
        });
    }

    function init_radio_buttons($el) {
        var cmd = $el.data('scpi');
        // Name is needed to make a proper radio group
        var name = 'radio_' + cmd;

        // The HTML structure looks like:
        // div.flex-col
        //   div.flex-row
        //     label
        //     input
        //   ...
        //   div.flex-row
        //     label
        //     input

        var outer_wrapper = $('<div>');
        outer_wrapper.addClass('flex-col');

        var options = $el[0].options;

        var radios = [];

        for (var i = 0; i < options.length; i++) {
            var text = options[i].text;
            var value = options[i].value;
            // ID is needed to make a clickable label
            var id = name + '_' + value;

            var label = $('<label>');
            label.attr('for', id);
            label.text(text);
            label.addClass('flex-fill');

            var radio = $('<input>');
            radio.attr('id', id);
            radio.attr('value', value);
            radio.attr('type', 'radio');
            radio.attr('name', name);
            radio.prop('indeterminate', true);

            radio.on('change', function () {
                scpi.set_value(cmd, $(this).val());
            });

            radios.push({
                radio: radio,
                value: value
            });

            var inner_wrapper = $('<div>');
            inner_wrapper.append(radio);
            inner_wrapper.append(label);
            inner_wrapper.addClass('flex-row');

            outer_wrapper.append(inner_wrapper);
        }

        $el.replaceWith(outer_wrapper);

        scpi.on(cmd, function (new_value) {
            radios.map(function (o) {
                o.radio.prop('checked', new_value === o.value);
            });
        });
    }

    $('[data-scpi]').each(function () {
        var that = $(this);
        var cmd = that.data('scpi');
        var tagname = that.prop('tagName').toLowerCase();

        if (tagname === 'span' || tagname === 'div') {
            scpi.on(cmd, function (value) {
                that.text(value);
            });
        }
        else if (tagname === 'button') {
            that.on('click', function () {
                scpi.send(cmd);
            });
        }
        else if (tagname === 'select') {
            that.val('');
            if (that.hasClass('scpi-text')) {
                init_select_text(that);
            }
            else if (that.hasClass('to-radio-buttons')) {
                init_radio_buttons(that);
            }
            else {
                that.on('change', function () {
                    scpi.set_value(cmd, that.val());
                });
                scpi.on(cmd, function (value) {
                    that.val(value);
                });
            }
        }
        else if (tagname === 'input') {
            var input_type = that.prop('type').toLowerCase();
            var setting_input_type = 'text';

            if (input_type === 'checkbox') {
                init_checkbox(that);
            }
            else {
                if (that.closest('.frequency').length > 0) {
                    that.prop('type', 'number');
                    that.setting_input('init', 'frequency');
                }
                else if (that.closest('.time').length > 0) {
                    that.prop('type', 'number');
                    that.setting_input('init', 'time');
                }
                else if (that.closest('.memory').length > 0) {
                    that.prop('type', 'number');
                    that.setting_input('init', 'memory');
                }
                else if (input_type === 'number') {
                    that.setting_input('init', 'number');
                }
                else if (input_type === 'text') {
                    that.setting_input('init', 'text');
                }
                else {
                    console.log(that);
                    throw 'Not sure what to do with data-scpi for this element';
                }

                scpi.on(cmd, function (value) {
                    that.setting_value(value);
                });

                that.on('commit', function (event, value) {
                    scpi.set_value(cmd, value);
                });
            }
        }
        else {
            console.log(that);
            throw 'Not sure what to do with data-scpi for this element';
        }
    });
}

function Tabs(scpi) {
    function tab_group_selector(val) {
        if (typeof val === 'undefined') {
            return '[data-tab-group]';
        }
        else {
            return '[data-tab-group="' + val + '"]';
        }
    }

    function tab(name) {
        return $('#setting-panel-tabs').children(tab_group_selector(name));
    }

    function tab_content(name) {
        return $('#settings-panel').children(tab_group_selector(name));
    }

    function tab_group_attr(el) {
        return $(el).attr('data-tab-group');
    }

    // Validate that every tab has at least one element in the right sidebar
    tab().each(function (i, el) {
        var group_name = tab_group_attr(el);
        if (tab_content(group_name).length == 0) {
            throw 'No elements for group ' + group_name;
        }
    });

    // Validate that every element in the right sidebar belongs to one tab group
    $('#settings-panel>*').each(function (i, el) {
        var group_name = tab_group_attr(el);
        if (typeof group_name === 'undefined') {
            console.log(el);
            throw 'Element does not have a group name';
        }

        if (tab(group_name).length == 0) {
            throw 'No tab for group ' + group_name;
        }
    });

    function callback_set() {
        var items = {};
        return {
            push: function (name, fn) {
                check_tab_exists(name);
                if (typeof fn !== 'function') {
                    throw 'Not a function: ' + fn;
                }
                items[name] = items[name] || [];
                items[name].push(fn);
            },
            call: function(name) {
                (items[name] || []).map(call);
            },
        };
    }

    var onfocus = callback_set();
    var onblur = callback_set();

    function set_active_tab(name) {
        tab().each(function () {
            $(this).attr('title', 'Show ' + $(this).text().trim() + ' Settings');
        });

        onblur.call(active_tab_name);

        if (name === active_tab_name) {
            onblur.call(name);
            active_tab_name = '';
            $(document.body).addClass('settings-collapsed');
            tab().removeClass('active');
            tab_content().addClass('hidden');
        }
        else {
            onfocus.call(name);
            active_tab_name = name;
            $(document.body).removeClass('settings-collapsed');
            tab().not(tab_group_selector(name)).removeClass('active');
            tab(name).addClass('active');
            tab_content().not(tab_group_selector(name)).addClass('hidden');
            tab_content(name).removeClass('hidden');

            tab(name).attr('title', 'Hide All Settings');

            setTimeout(function () {
                var elems = tab_content(name).
                    find('input, select, textarea, button').
                    filter(':enabled');

                if (elems.length > 0) {
                    elems[0].focus()
                }
            }, 10);
        }
    }

    set_active_tab();

    tab().on('click', function () {
        set_active_tab(tab_group_attr(this));
    });

    var storage = poll_json_local_storage('selected-tab', { name: 'frequency' });

    storage.onrestore(function (stored) {
        if (stored) {
            set_active_tab(stored.name);
        }
    });

    storage.onsave(function () {
        return { name: active_tab_name };
    });

    scpi.onconnect(function () {
        scpi.query_and_listen('DEB:SYST:LOCK:LEV?', function(val) {
            if ($.trim(val) === "2") {
                tab('debug').removeClass('hidden');
            }
            else {
                tab('debug').addClass('hidden');
            }
        });
    });

    function check_tab_exists(name) {
        if (tab(name).length < 1 || tab_content(name).length < 1) {
            throw 'Invalid tab name ' + name;
        }
    }

    return {
        onfocus: function (tabname, callback) {
            onfocus.push(tabname, callback);
            if (tabname === active_tab_name) {
                onfocus.call(tabname);
            }
        },
        onblur: function (tabname, callback) {
            onblur.push(tabname, callback);
        },
    };
}

function setup_app_version(client) {
    client.onconnect(function() {
        client.send('app-version', function (val) {
            $('#app-version').html(val);
        });
    });
}

function setup_storage_info(client) {
    client.onconnect(function () {
        client.join('storage-directory', function (msg) {

           var available_msuses = msg.usbMedia.slice(0);

           storage_device_changes(available_msuses);
        });
    });
}

function setup_scpi_console(client) {
    var editing_line = '';

    var scpi_command_history = (function () {
        var commands = [''];
        var index = -1;

        return {
            prev: function () {
                index = Math.max(index - 1, 0);
                return commands[index];
            },
            next: function () {
                // Out of bounds intentionally
                index = Math.min(index + 1, commands.length);
                if (index >= commands.length) {
                    return editing_line;
                }
                else {
                    return commands[index];
                }
            },
            push: function (cmd) {
                if (cmd !== commands[commands.length - 1]) {
                    commands.push(cmd);
                }
                index = commands.length; // Out of bounds intentionally
            }
        };
    })();

    function send_scpi_cmd(cmd) {
        if (String(cmd).length !== 0) {
           client.send('scpi', cmd, function(msg) {
              if (typeof msg === 'string' ) {
                  append_rejectLog(msg);
                  scroll_log();
              }
            });
            scpi_command_history.push(cmd);
            cmd_el.val('');
            editing_line = '';
        }
    }

    $('#send-scpi-cmd').on('click', function () {
        send_scpi_cmd(cmd_el.val());
    });

    var cmd_el = $('#scpi-command-entry');
    cmd_el.on('keyup', function (event) {
        var cmd = cmd_el.val();
        switch (event.which || event.keyCode) {
            case 38: // Up arrow
                cmd_el.val(scpi_command_history.prev());
                break;
            case 40: // Down arrow
                cmd_el.val(scpi_command_history.next());
                break;
            case 13:
                send_scpi_cmd(cmd);
                break;
            default:
                editing_line = cmd;
                break;
        }
    });

    var $log = $('#scpi-log');
    function scroll_log() {
        $log.scrollTop($log.prop('scrollHeight'));
    }

    function append_rejectLog(msg) {
      var error_el = $('<div class=scpi-error>');
      error_el.text(msg)
              .addClass('emphasis');
      $log.append(error_el);
    }

    function append_scpi(msg) {
        var req_el = $('<div class=scpi-command>');

        req_el.text(msg.command);
        $log.append(req_el);
        scpi_command_history.push(msg.command);


        (msg.errors || []).map(function (error) {
            var text;
            var error_el = $('<div class=scpi-error>');

            if (typeof error.num !== 'undefined') {
                text = 'Error ' + error.num;
            }
            else {
                text = 'Error';
            }
            error_el.text(text + ': ' + error.description);
            $log.append(error_el);
        });

        if ((msg.response || '').length > 0) {
            var resp_el = $('<div class=scpi-response>');
            resp_el.text(msg.response);
            $log.append(resp_el);
        }
    }

    client.onconnect(function () {
        $log.empty();

        client.join('scpi-log', function (msg) {
            if (!msg.quiet) {
                append_scpi(msg);
                scroll_log();
            }
        });
    });
}

function select_tab_inputs(tabnames) {
    var result = null;

    for (var index = 0; index < tabnames.length; index++) {
        var tabname = tabnames[index];
        var elements =  $('.setting-group[data-tab-group="'+ tabname +'"] input,' +
                          '.setting-group[data-tab-group="'+ tabname +'"] button,' +
                          '.setting-group[data-tab-group="'+ tabname +'"] select');
        if (result !== null) {
            result = result.add(elements);
        }
        else {
            result = elements;
        }
    }

    return result;
}

function setup_overheating_msg(client) {
    var msg_displayed = false;
    var modal = modal_overlay('#overheating');
    var tabs_to_disable = ["frequency", "amplitude-bw", "sweep", "meas-1",
                           "markers", "limits","iq-capture"];

    var disabled_elems = select_tab_inputs(tabs_to_disable);

    function on_overheating() {
        $('#temperature-status').html('&middot; Unit Overheating')
                             .addClass('overheating');
        disabled_elems.prop('disabled', true);
    }

    function on_normal_temperature() {
        $('#temperature-status').html('').removeClass('overheating');
        disabled_elems.prop('disabled', false);
    }

    client.onconnect(function() {
        client.join('overheat-status', function (val) {
            if (val) {
                if (!msg_displayed) {
                    on_overheating();
                    modal.show();
                    msg_displayed = true;
                }
            }
            else {
                on_normal_temperature();
                modal.hide();
                msg_displayed = false;
            }
        });
    });
}

function setup_disconnect_msg(client) {
    var overlay = modal_overlay('#disconnected');
    var was_connected = true;

    var disabled_elems = $('.setting-group input,' +
                           '.setting-group button,' +
                           '.setting-group select');

    disabled_elems.prop('disabled', true);

    client.onconnect(function () {
        was_connected = true;
        overlay.hide();
        $('#connection-status').html('').removeClass('disconnected');
        disabled_elems.prop('disabled', false);
    });

    function ondisconnect() {
        $('#connection-status').html('&middot; Disconnected').
            addClass('disconnected');
        disabled_elems.prop('disabled', true);
        if (was_connected) {
            overlay.show()
        }
        was_connected = false;
    }

    setTimeout(function () {
        // Give the client a chance to connect the first time
        if (!client.connected()) {
            ondisconnect();
        }

        client.ondisconnect(function () {
            ondisconnect();
        })
    }, 2000);
}

var spectrogram_visible = (function init_spectrogram_visible() {
    if (typeof poll_json_local_storage === 'undefined') {
        return function () { return false };
    }
    var $el = $('#show-spectrogram');
    var storage = poll_json_local_storage('show-spectrogram', true);

    var visible;
    $el.on('change', function () {
        visible = $el.is(':checked');
    }).trigger('change');

    $el.ready(function(){
        visible = $el.is(':checked');
    });

    function spectrogram_visible() {
        return visible
    }

    storage.onsave(spectrogram_visible);

    storage.onrestore(function (checked) {
        $el.prop('checked', checked);
    });

    return spectrogram_visible;
})();

function setup_reboot(scpi) {
    $('#reboot').on('click', function () {
        scpi.send('*rst');
    })
}

function update_trace_refresh_rate(trace_data){

    var is_checked = $('#set-manual-refresh-rate').is(':checked');
    var value_in_secs = parseFloat($('#trace-refresh-rate-value').val());
    var rate_in_ms = (value_in_secs*1000);

    trace_data.set_trace_refresh_parameters({manual_refresh:is_checked,refresh_rate_in_ms:rate_in_ms});
}

function setup_trace_refresh_rate(trace_data){

    var $el = $('#trace-refresh-rate-value');

    $el.setting_input('init', 'number');

    var prev_valid_value;

    $el.on('focus',function(){
        prev_valid_value = this.value;

    }).on('commit', function (_, val) {

        if(val > 60 || val < 1)//max value is 60 secs and min value is 1 sec
            $el.setting_input('value', prev_valid_value);

        update_trace_refresh_rate(trace_data);
    });

    $el.setting_value($el.attr('value'));

    $('#set-manual-refresh-rate').change(function(){
        update_trace_refresh_rate(trace_data);
    });
}

function setup_option_upgrade(client) {
    var modal = modal_overlay('#option-modal');
    var status_text = $('#license-status-text');
    var status_details = $('#license-status-details');
    var reboot_button = $('#option-reboot-button');
    var license_path = $('#license-file-path');
    var upload_button = $('#upload-license-button');

    $('#option-upgrade-button').on('click', function() {
        license_path.attr('value', null).data('file', null);
        status_text.text('...');
        status_text.css('color', '');
        status_details.text('');
        modal.show();
        reboot_button.prop('disabled', true);
        upload_button.prop('disabled', true);
    }).val('');

	$('#download-configuration-button').on('click', function() {
		client.send('scpi-quiet', 'syst:opt:conf?', function (response) {
			var idn = instrument_idn($("#firmware-info").text());
			var name = idn.model + " " + idn.serial + ".config";
			var data = null;
			if (!response.errors) {
				data = response.response;
				saveAs(new Blob([data], {type: "text/plain;charset=utf-8"}), name);
			}
		});
	}).val('');

	$('#license-file-choose-button').on('click', function() {
		$('#license-file-choose-input').attr('accept', '.license').click();
	}).val('');

	$('#license-file-choose-input').on('change', function(e) {
		var file = e.target.files[0];
		$('#license-file-path').attr('value', (file) ? $(this).val() : null).data('file', file);

        if (file) {
            upload_button.prop('disabled', false);
        }
        else {
            upload_button.prop('disabled', true);
        }
	});

	function upload(file, callback) {
		var reader = new FileReader();
		reader.onload = function(e) {
			var data = e.target.result;

            // Sanity check, data should always be a quoted string
            if (data.length <= 2 || data[0] !== '"' || data[data.length - 1] !== '"') {
                callback({'errors' : [{'description' : ''}]});
            }
            else {
                client.send('scpi-quiet', 'syst:opt:upgr ' + data, function(response) {
                    callback(response);
                });
            }
		};
        reader.readAsBinaryString(file);
	}

    upload_button.on('click', function() {
        var file = license_path.data('file');
        if (file) {
            status_details.text('');
            status_text.text('Installing...');
            status_text.css('color', 'blue');

            upload(file, function(response) {
                if (!response.errors) {
                    status_text.text('Successful!');
                    status_text.css('color', 'green');
                    status_details.text('');
                    reboot_button.prop('disabled', false);
                }
                else {
                    status_text.text('Failed!');
                    status_text.css('color', 'red');
                    reboot_button.prop('disabled', true);

                    var code = response.errors[0].num;
                    var details = response.errors[0].details;

                    if (code == -200) {
                        if (details) {
                            status_details.text(details.replace(/.*- /i, " "));
                        }
                    }
                    else {
                        status_details.text('License file has invalid format');
                    }

                }
            });
        }
    }).val('');

    reboot_button.on('click', function () {
        client.send('scpi-quiet', '*rst');
    }).val('');
}

function setup_firmware_update(client, scpi) {
    $('#firmware-update-button').on('click', function () {
        var name = String($('#firmware-update').find(':selected').data('name'));
        if (name.length > 0) {
            scpi.send('syst:firm:upd:rem "' + name + '"');
            $('#firmware-update-msg').show();
        }

        $(this).val('');
    }).val('');

    var check_modal = modal_overlay('#fwupdate-check');
    var prompt_modal = modal_overlay('#fwupdate-prompt');
    var up_to_date_modal = modal_overlay('#fwupdate-up-to-date');

    $('#fwupdate-prompt-button').on('click', function () {
        var name = String($('#fwupdate-prompt-package').html());
        if (name.length > 0) {
            scpi.send('syst:firm:upd:rem "' + name + '"');
            prompt_modal.hide();
            $('#firmware-update-msg').show();
        }
    }).val('');

    $('#fwupdate-reinstall-button').on('click', function () {
        var name = String($('#fwupdate-reinstall-package').html());
        if (name.length > 0) {
            scpi.send('syst:firm:upd:rem "' + name + '"');
            up_to_date_modal.hide();
            $('#firmware-update-msg').show();
        }
    }).val('');

    $('#firmware-update-check-latest').on('click', function() {
        check_modal.show();
        scpi.send_quiet('syst:firm:upd:rem:lat?', function(response) {
            var values = response.trim().split(',');
            var package_name = values[0];
            var update_status = values[1];
            $('#fwupdate-prompt-package').html(package_name);
            $('#fwupdate-reinstall-package').html(package_name);

            if (package_name.length > 0) {
                check_modal.hide();

                if (update_status === '1') {
                    prompt_modal.show();
                }
                else if (update_status === '0') {
                    up_to_date_modal.show();
                }
            }
        });
    }).val('');

    var fwupdate_adv_el = $('#firmware-update-advanced-checkbox');

    fwupdate_adv_el.on('change', function () {
        var checked = fwupdate_adv_el.prop('checked');
        if (checked)
        {
            $('#firmware-update-check-latest').hide();
            $('#firmware-update-button').show();
            $('#firmware-update').show();
        }
        else
        {
            $('#firmware-update-check-latest').show();
            $('#firmware-update-button').hide();
            $('#firmware-update').hide();
        }

    }).val('');
    $('#firmware-update-check-latest').show();
    $('#firmware-update-button').hide();
    $('#firmware-update').hide();

    var fwupdate_el = $('#firmware-update');
    var progress_modal = modal_overlay('#fwupdate-in-progress');

    client.onconnect(function () {
        progress_modal.hide();
        check_modal.hide();
        prompt_modal.hide();
        up_to_date_modal.hide();

        fwupdate_el.empty();
        client.send('fw-update-sources', function (msg) {
            (msg || []).map(function (source) {
                var resp_el = $('<option value=yes></option>');
                resp_el.text(source.name);
                resp_el.data('name', source.name);
                resp_el.attr('recommended', source.recommended);
                if (source.recommended) {
                    resp_el.attr('selected', 'selected');
                    }
                $('#firmware-update').append(resp_el);
            })
        });

        client.join('fwupdate', function (msg) {
            console.log("Downloading update from " + msg.url);

            $('.fwupdate-message').text(msg.message || 'Unknown error');

            if (msg.what === 'started') {
                progress_modal.show();
            }
            else if (msg.what === 'error') {
                progress_modal.hide();
            }
        });
    });
}

function $scpi(cmd) {
    var sel = '[data-scpi="' + cmd + '"]';
    var ret = $(sel);
    if (ret.length !== 1) {
        throw 'Bad number of elements (' + ret.length +
              ') having selector ' + sel;
    }
    return ret;
}

function setup_rl_buttons(scpi) {
    /* I am loathe to have so many special cases. If you can think of a
     * simpler or better way to implement this, please do
     */

    var rl_cmd = 'disp:wind:trac:y:scal:rlev';
    var off_cmd = 'disp:wind:trac:y:scal:rlev:offs';
    var scale_cmd = 'disp:wind:trac:y:scal:pdiv';

    var $rl = $scpi(rl_cmd);
    var $off = $scpi(off_cmd);

    function get(cmd) {
        return Number($scpi(cmd).val());
    }

    function set_off(x) {
        scpi.set_value(off_cmd, x);
    }

    function set_rl(x) {
        scpi.set_value(rl_cmd, x );
    }

    $('#rl-up').on('click', function () {
        set_rl(get(rl_cmd) + get(scale_cmd));
    });

    $('#rl-down').on('click', function () {
        set_rl(get(rl_cmd) - get(scale_cmd));
    });

    $('#rl-offset-up').on('click', function () {
        set_off(get(off_cmd) + get(scale_cmd));
    });

    $('#rl-offset-down').on('click', function () {
        set_off(get(off_cmd) - get(scale_cmd));
    });

}

function setup_amplitude_span(trace_data, spectrum) {

    var $el = $("input[name=amplitude-scale]");

    $el.on('commit', function (_, scale) {
        set_amplitude_span(scale, spectrum, trace_data);
    });
}

function setup_show_firmware_info(scpi) {
    scpi.onconnect(function () {
        scpi.send_quiet('*idn?',function(val){
            $("#firmware-info").text(val);
            var idn = instrument_idn(val);
            $("#installed-options").text((idn.options.length == 0) ? 'None' : idn.options);
        });
    });
}

function setup_unit_name_info(scpi) {
    scpi.onconnect(function () {
        scpi.send_quiet('syst:unit:nam?',function(val){
            $("#unit-name").text(val);
          });
    });
}

function setup_show_nfsmount_info(scpi) {
    scpi.onconnect(function () {
        scpi.query_and_listen('SYSTem:COMMunicate:LAN:NFS:MOUNt:LIST:REMHost?',function(val){
        $("#nfsmount-info").text(val);
        if((val).trim() === "None") {
            $('#nfsmount-label').hide();
            $('#nfsmount-info').hide();
        }
        else {
            $('#nfsmount-label').show();
            $('#nfsmount-info').show();
        }
        });
    });
}

function setup_scpi_script(scpi) {
    $('#scpi_script').val('').on('change', function (evt){
        for (var i = 0; i < evt.target.files.length; i++) {
            var f = evt.target.files[i];
            if (f.name.match(/\.txt$/i)) {
                var reader = new FileReader();

                reader.onload = function (){
                    var lines = this.result.split('\n');

                    function send_next_line() {
                        if (lines.length > 0)
                        scpi.send(lines.shift(), send_next_line);
                    }

                    send_next_line();
                };
                reader.readAsText(f);
            }
            else {
                alert("The SCPI script must have the .txt extension");
            }
        }

        $('#scpi_script').val('');
    });
}

function send_static_ethernet_settings(scpi) {
    var static_ip = $('#static-ip').val();
    var static_subnet = $('#static-subnet').val();
    var static_gateway = $('#static-gateway').val();

    if(static_ip.length > 0 && static_subnet.length > 0 && static_gateway.length > 0)
    {
        var ethernet_settings = ('"' + static_ip + '","' +
                                       static_gateway + '","' +
                                       static_subnet + '"');
        scpi.send("syst:comm:lan:conf"+" "+ethernet_settings);
    }
}

function setup_dhcp(scpi) {
    var $el = $('#ip-address-allocation');
        var static_group = $('.static-ethernet');
        var scpi_cmd = 'syst:comm:lan:dhcp?';

        scpi.onconnect(function () {
            scpi.send_quiet(scpi_cmd, function (scpi_value) {
                scpi_value = scpi_value.toLowerCase();
                scpi_value = scpi_value.replace('\n', '');
                var selected;

                if (scpi_value === 'on') {
                    static_group.addClass('hidden');
                    selected = 'dhcp';
                }
                else {
                    static_group.removeClass('hidden');
                    selected = 'static';
                }

                $("#ip-address-allocation").val(selected);
            });
        });

        $el.on('change', function () {
            var val = $el.val();

            if (val === 'dhcp') {
                static_group.addClass('hidden');
            }
            else if (val === 'static') {
                static_group.removeClass('hidden');
            }
        });
}

function setup_ethernet(scpi) {
    $('#ethernet-apply').on('click', function () {
        var ethernet_configuration = $('#ip-address-allocation').val();
        if(ethernet_configuration === 'dhcp') {
            scpi.send("syst:comm:lan:dhcp on");
        }
        else {
            send_static_ethernet_settings(scpi);
        }
        scpi.send_quiet("syst:comm:lan:conf?");
        scpi.send_quiet("syst:comm:lan:conf:curr?");
        scpi.send_quiet("syst:comm:lan:dhcp?");
    });
}

function get_ethernet_startup_settings(scpi) {
    scpi.onconnect(function () {
        scpi.send_quiet("syst:comm:lan:conf?", function (response) {
                var ethernet_settings = response.split(",");
                if(ethernet_settings.length === 3)
                {
                    $('#static-ip').val(ethernet_settings[0]);
                    $('#static-gateway').val(ethernet_settings[1]);
                    $('#static-subnet').val(ethernet_settings[2]);
                }
        });
    });
}

function get_current_ethernet_startup_settings(scpi) {
    scpi.onconnect(function () {
        scpi.send_quiet("syst:comm:lan:conf:curr?", function (response) {
                var ethernet_settings = response.split(",");
                if(ethernet_settings.length === 3)
                {
                    $('#current-ip').text(ethernet_settings[0]);
                    $('#current-ip').attr('href',
                        location.protocol + '//' +
                        ethernet_settings[0]);
                    $('#current-gateway').text(ethernet_settings[1]);
                    $('#current-subnet').text(ethernet_settings[2]);
                }
        });
    });
}

function setup_dns(scpi) {
    $('#dns-apply').on('click', function () {
        var dns_host = $('#dns-host').val();
        var dns_username = $('#dns-username').val();
        var dns_password = $('#dns-password').val();

        if (dns_host.length > 0 && dns_username.length > 0 && dns_password.length > 0)
        {
            var dns_settings = ('"' + dns_host + '","' +
                                      dns_username + '","' +
                                      dns_password + '"');
            scpi.send("syst:comm:lan:dns"+" "+dns_settings);
            scpi.send_quiet("syst:comm:lan:dns?");
        }
    });
}

function get_dns_startup_settings(scpi,client) {
    scpi.onconnect(function () {
        scpi.send_quiet("syst:comm:lan:dns?", function (response) {
                var dns_settings = response.split(",");
                if(dns_settings.length === 2)
                {
                    $('#dns-host').val(dns_settings[0]);
                    $('#dns-username').val(dns_settings[1]);
                    $('#dns-password').val("");
                }
        });
    });
}

function setup_report_bug(scpi) {
    scpi.onconnect(function () {
        scpi.send_quiet('*idn?', function (build_tag) {
            var description = '\n\n\n\nBuild: ' + build_tag;

            $('#report-bug-jira').attr('href',
                'http://jira/secure/CreateIssueDetails!init.jspa?' +
                $.param({
                    pid: 10000, // TODO hard coded ID
                    issuetype: 1,
                    priority: 3,
                    description: description
                }))
            .attr('target', '_blank');
        });
    });
}

var font_size = (function setup_font_size() {
    var size = '10pt';
    if (typeof poll_json_local_storage === 'undefined') {
        return function () { return size; };
    }
    var size_px = $('body').css('font-size');
    var storage = poll_json_local_storage('font-size', size);

    function set_font_size(x) {
        size = x;
        $('body').css('font-size', x);
        size_px = $('body').css('font-size');
        $('#text-zoom').val(x);
    }

    storage.onrestore(set_font_size);
    storage.onsave(function () {
        return size;
    });

    $('#text-zoom').on('change', function () {
        set_font_size($(this).val());
    });

    return function font_size() {
        return size_px;
    };
})();

function setup_display_filters() {
    var filters = {
        'display-brightness': 'brightness',
        'display-contrast': 'contrast',
        'display-saturation': 'saturate',
    };

    function set_filters() {
        var val = '';
        for (var id in filters) {
            var $el = $('#' + id);
            val += filters[id] + '(' + $el.val() + '%) ';
        }
        var $elems = $('#spectrum, #setting-panel-tabs');
        $elems.css('filter', val)
        $elems.css('-moz-filter', val)
        $elems.css('-webkit-filter', val)
    }

    for (var id in filters) {
        $('#' + id).on('input change', set_filters).trigger('change');
    }

    $('#reset-display').on('click', function () {
        for (var id in filters) {
            $('#' + id).val(100).trigger('change');
        }
    });
}

function setup_debug_input(client, scpi) {
    scpi.onconnect(function () {
        $('#lo-freq-input').on('keydown', function (e) {
            if(e.keyCode == 13) {
                var lo = $('#lo-freq-select').val();
                var freq = $('#lo-freq-input').val();
                scpi.send('deb:lo:freq ' + lo + ', ' + freq);
            }
        });
    });
}

function setup_average_count(scpi) {
    var $avg_count = $('[data-scpi="sens:aver:coun"]');
    var enabled_for = { aver: true, rmax: true, rmin: true };

    scpi.on('sens:aver:type', function (val) {
        $avg_count.prop('disabled', !enabled_for[val.toLowerCase()]);
    });
}

function setup_preselector(scpi) {
    scpi.onconnect(function () {
        scpi.send_quiet('*idn?',function(val){
            var $preselector = $('[data-scpi="sens:freq:band:mode"]');
            const PRESELECTOR_OPT_NUMBER = '410';
            var idn = instrument_idn(val);
            var modelOptions = idn.options;

            if (!modelOptions.includes(PRESELECTOR_OPT_NUMBER)) {
                $preselector.prop('disabled', true);
            }
        });
    });
}

function setup_obw_method(scpi) {
    var $percent = $('[data-scpi="sens:obw:perc"]');
    var $xdb = $('[data-scpi="sens:obw:xdb"]');

    function nearest_setting($el) {
        var ret = $el.closest('.setting');
        if (ret.length === 0) return $el;
        else return ret;
    }

    scpi.on('sens:obw:meth', function (val) {
        if (val === 'XDB') {
            $xdb.prop('disabled', false);
            nearest_setting($xdb).removeClass('hidden');
            nearest_setting($percent).addClass('hidden');
        }
        else if (val === 'PERC') {
            $percent.prop('disabled', false);
            nearest_setting($percent).removeClass('hidden');
            nearest_setting($xdb).addClass('hidden');
        }
        else {
            $percent.prop('disabled', true);
            $xdb.prop('disabled', true);
            nearest_setting($xdb).addClass('hidden');
            nearest_setting($percent).addClass('hidden');
        }
    });
}

function setup_limit_checking(scpi) {

    var $enable_limit_checking = $('[data-scpi="calc:lim:stat"]');
    var $set_max_hold = $('[data-scpi="calc:lim:upp:stat"]');
    var $set_min_hold = $('[data-scpi="calc:lim:low:stat"]');
    var $limit_save_on_event = $('[data-scpi="mmem:stoe:lim"]');

    var update_checkbox = function ($element, response) {
        response = response.replace('\n','');

        var isChecked = (response === '1' ? true : false);
        $element.prop('indeterminate', false);
        $element.prop('checked', isChecked);

    };
    scpi.onconnect(function () {
        scpi.send_quiet('calc:lim:stat?', function(response) {
            update_checkbox($enable_limit_checking, response);
        });

        scpi.send_quiet('calc:lim:upp:stat?', function(response) {
            update_checkbox($set_max_hold, response);
        });

        scpi.send_quiet('calc:lim:low:stat?', function(response) {
            update_checkbox($set_min_hold, response);
        });
    });

    /*this keeps the trace check, upper and lower status in sync when limit state
    is changed*/
    $enable_limit_checking.on('change', function(){

        //get the new value of limit state
        var val = $enable_limit_checking.prop('checked');
        val = (val == false ? '0' : '1');

        //send scpi to backend
        scpi.set_value('calc:lim:upp:stat', val);
        scpi.set_value('calc:lim:low:stat', val);
        scpi.set_value('calc:lim:trac:check', val);

        //query them back to keep gui in sync
        scpi.send_quiet('calc:lim:upp:stat?', function(response) {
            update_checkbox($set_max_hold, response);
        });

        scpi.send_quiet('calc:lim:low:stat?', function(response) {
            update_checkbox($set_min_hold, response);
        });
    });

}

// sorry for the global
var global_gps_description = '';

function setup_gps_status(client) {
    var $icon = $('#gps-info');

    function sigfigs(val, digits) {
        var mult = Math.pow(10, digits);
        return Math.round(val * mult) / mult;
    }

    client.onconnect(function () {
        client.join('gps', function (info) {
            if (info.good_fix) {
                $icon.addClass('gps-good-fix');
                $icon.prop('title', 'Good GPS Fix');
                $('#gps-fix').text('Good fix');
                var position = info.latitude.toFixed(6) + '&deg;' +
                               ', ' +
                               info.longitude.toFixed(6) + '&deg;'
                $('#gps-position').html(position);
                $('#gps-altitude').text(sigfigs(info.altitude, 4) + ' m');
                $('#gps-satellites').text(info.num_satellites);
                $('.gps-info').show();

                global_gps_description = position;
            }
            else {
                $icon.removeClass('gps-good-fix');
                $icon.prop('title', 'No GPS Fix');
                $('#gps-fix').text('None');
                $('#gps-position').text('');
                $('#gps-altitude').text('');
                $('#gps-satellites').text('');
                $('.gps-info').hide();

                if (info.state) {
                    global_gps_description = 'No GPS Fix';
                }
                else {
                    global_gps_description = '';
                }
            }

            $('#gps-enabled').text(info.state ? 'Yes' : 'No');
        });
    });
}

var get_spectrum_metadata;
var iq_metadata_snapshot;

function init_get_metadata(trace_data) {
    function getter(scpi_cmd) {
        return function (scpi) {
            var ret = scpi.get(scpi_cmd);
            if (typeof ret === 'undefined') return '';
            return ret;
        }
    }

    function verbose_getter(scpi_cmd) {
        var $el = $('select[data-scpi="' + scpi_cmd + '"]');

        if ($el.length === 0) {
            throw 'No select element for cmd ' + scpi_cmd;
        }

        return function (scpi) {
            var val = scpi.get(scpi_cmd);

            if (typeof val === 'undefined') return '';
            var opts = $el[0].options;
            for (var i = 0; i < opts.length; i++) {
                if (opts[i].value === val) {
                    return opts[i].text;
                }
            }
            return val || '';
        }
    }

    var ref_level = getter('disp:wind:trac:y:scal:rlev');
    var atten = getter('sens:pow:rf:att');
    var det = verbose_getter('trac1:det:func');
    var rbw = verbose_getter('sens:band:res');
    var vbw = verbose_getter('sens:band:vid');
    var sweep_mode = verbose_getter('swe:mode');
    var preamp = getter('sens:pow:rf:gain:stat');
    var cont = verbose_getter('init:cont');
    var points = getter('disp:poin');
    var units = verbose_getter('unit:pow');

    get_spectrum_metadata = function(scpi) {
        return {
            'Ref Lvl': ref_level(scpi) + ' ' + units(scpi),
            'Inp Atten': atten(scpi) + ' dB',
            'Preamp': preamp(scpi) === '1' ? 'On' : 'Off',
            'Detection': det(scpi),
            'RBW': rbw(scpi),
            'VBW': vbw(scpi),
            'Sweep': [
                sweep_mode(scpi),
                cont(scpi),
                trace_data.pretty_sweep_speed(),
                '' + points(scpi) + ' Points'],
        };
    }

    var center_frequency = getter('sens:freq:cent');
    var iq_bandwidth = getter('sens:iq:band');
    var iq_bits = getter('sens:iq:bits');
    var iq_ts = getter('sens:iq:time');
    var iq_length = getter('sens:iq:leng');

    iq_metadata_snapshot = function(scpi) {
        return {
            'Ref Lvl': ref_level(scpi) + ' ' + units(scpi),
            'Inp Atten': atten(scpi) + ' dB',
            'Preamp': preamp(scpi) === '1' ? 'On' : 'Off',
            'Center Freq': format_freq(center_frequency(scpi)),
            'I/Q Capture': [
                format_freq(iq_bandwidth(scpi)),
                format_microseconds(iq_length(scpi)),
                iq_bits(scpi) + ' Bits',
                iq_ts(scpi) === '1' ? 'Timestamped' : 'No Timestamps',
            ],
        };
    }
}

function instrument_idn(idn) {
    if (typeof idn === 'undefined') {
        return {};
    }
    var toks = idn.split(',');
    var anritsu = toks.shift();
    var modelOptions = toks.shift() || '';
    var res = modelOptions.split('/');
    var model = res.shift();
    var options = res.join('/');
    var version = toks.pop();
    var serial = toks.pop();

    return {
        model: model || '',
        options: options || '',
        serial: serial || '',
        version: version || '',
    };
}

function setup_span_up_down(scpi) {
    var cmd = 'sens:freq:span';

    $('#span-up').on('click', function () {
        scpi.set_value(cmd, snap.up(parseFloat(scpi.get(cmd))));
    });

    $('#span-down').on('click', function () {
        scpi.set_value(cmd, snap.down(parseFloat(scpi.get(cmd))));
    });
}

function setup_iq_capture_result(client, set_iq_data_callback) {
    var max_raw_size = 4096;
    var spinner = document.getElementById('spinner');
    var message = document.getElementById('progress-message');

    function get_iq_data(onload) {
        var xhr = new XMLHttpRequest();

        function check_status() {
            if ((xhr.status > 0 && xhr.status < 200) || xhr.status > 300) {
                message.innerHTML += '<br>Resource unavailable. Is somebody' +
                                     ' else using the SCPI port?';
                message.style.color = 'red';
            }
        }

        xhr.addEventListener('progress', function (event) {
            check_status();
            if (xhr.response.length >= max_raw_size) {
                onload(xhr.response);
                xhr.abort();
            }
        })

        xhr.addEventListener('load', function (event) {
            onload(xhr.response);
            check_status();
        });

        function fail(event) {
            check_status();
        }

        xhr.addEventListener('error', fail);
        xhr.addEventListener('abort', fail);

        xhr.open('GET', '/traceIqData.csv');
        xhr.setRequestHeader('Range', 'bytes=0-' + max_raw_size);
        xhr.responseType = "text";
        xhr.timeout = 10000;
        xhr.send();
    }

    function onload(response) {
        $('#iq-data-response-csv').val(response.trim().slice(0, max_raw_size));

        var rows = response.split('\r\n').slice(0,-2).map(function (line) {
            var pair = line.split(',');
            return {
                i: parseInt(pair[0]),
                q: parseInt(pair[1]),
            };
        }).filter(function (pair) {
            return !isNaN(pair.i) && !isNaN(pair.q);
        });

        set_iq_data_callback(rows);
    }

    function handle_capture_result() {
        client.join('iq-capture-result', function (val) {
             spinner.style.visibility = "hidden";
             message.style.color = 'red';
             var oldMessageInnerHtml = message.innerHTML;

             if(val == 5)
             {
                message.innerHTML = "IQ Capture Done Successfully";
                message.style.color = 'green';
                get_iq_data(onload);
                client.leave('iq-capture-result');
             }
             else if(val == 4)
             {
               message.innerHTML = "Host aborted capture";
               client.leave('iq-capture-result');
             }
             else if(val == 3)
             {
               message.innerHTML = "Over heat during capture capture paused";
               client.leave('iq-capture-result');
             }
             else if(val == 2)
             {
               message.innerHTML = "Over power during capture capture paused";
               client.leave('iq-capture-result');
             }
             else if(val == 1)
             {
               message.innerHTML = "Reference change event occured";
             }
             else if(val == 0)
             {
               message.innerHTML = "Paused capture was resumed";
               message.style.color = 'green';
             }

             message.innerHTML = oldMessageInnerHtml + '<br>' + message.innerHTML;
        });
    };

    $(function () {
        $('[data-scpi="meas:iq:capt"]').on('click', function () {
           message.innerHTML = "Starting IQ Capture";
           spinner.style.visibility = 'visible';
            $('#iq-data-response-csv').val('');
            set_iq_data_callback([]);
           handle_capture_result();
        });
    });
}

function LimitAlarm(client) {
    var audio = document.getElementById('limit-alarm');
    var $enable = $('#limit-alarm-enable');
    var $volume = $('#limit-alarm-volume');
    var last_fail_time = 0;

    client.onconnect(function () {
        client.join('limitFailure', function () {
            if ($enable.is(':checked')) {
                audio.volume = parseFloat($volume.val()) / 100;
                audio.play();
            }
            last_fail_time = Date.now();
        });
    });

    return {
        is_alarming: function () {
            return last_fail_time + 2000 > Date.now();
        },
    }
}

// holds the first found storage location name, usually "Internal"
var firstLocation;

var fileDialogs = {};
fileDialogs["#file-save-location"] = modal_overlay('#file-save-dialog');
fileDialogs["#file-recall-location"] = modal_overlay('#file-recall-dialog');
fileDialogs["#file-copy-from-location"] = modal_overlay('#file-copy-from-dialog');
fileDialogs["#file-copy-to-location"] = modal_overlay('#file-copy-to-dialog');
fileDialogs["#file-delete-location"] = modal_overlay('#file-delete-dialog');
//save_on_event_dialog = modal_overlay('#save-on-event-dialog');
fileDialogs["#save-on-event-location-cms"] = modal_overlay('#save-on-event-dialog');
fileDialogs["#save-on-event-location-limit"] = modal_overlay('#save-on-event-dialog');


function setup_file(client, scpi) {
    scpi.onconnect(function() {
        updateListofDevices();
        disableExitRecallMeasurement(true);
    });

    function download(path, location, name, callback) {
        client.send('scpi-quiet', 'mmem:data? "' + path + '","' + location + '"', function (response) {
            var data = null;
            if (!response.errors) {
                data = response.response.slice(parseInt(response.response.charAt(1)) + 2);
                saveAs(new Blob([data], {type: "text/plain;charset=utf-8"}), name);
            }
            typeof callback !== 'function' || callback(data, response);
        });
    }

    function upload(file, path, location, callback) {
        var reader = new FileReader();
        reader.onload = function(e) {
            var data = e.target.result;
            var size = data.length.toString();
            data = '#' + size.length + size + data;
            client.send('scpi-quiet', 'mmem:data "' + path + '","' + location + '",' + data, function (response) {
                typeof callback !== 'function' || callback(response);
            });
        };
        reader.readAsBinaryString(file);
    }

    var storeCommandPrefices = {
        'stp': 'mmem:stor:stat 0,"',
        'rsm': 'mmem:stor:trac "","',
        'lim': 'mmem:stor:lim "","'
    };

    function save(type, path, location, callback) {
        var commandPrefix = storeCommandPrefices[type];
        if (commandPrefix) {
            client.send('scpi-quiet', commandPrefix + path + '","' + location + '"', function (response) {
                typeof callback !== 'function'|| callback(response);
            });
        }
    }

    var loadCommandPrefices = {
        'stp': 'mmem:load:stat 0,"',
        'rsm': 'mmem:load:trac "","',
        'lim': 'mmem:load:lim "","'
    };

    function recall(type, path, location, callback) {
        if (!type) {
            var i = path.lastIndexOf('.');
            if (i !== -1) {
                type = path.slice(i + 1).toLowerCase();
            }
        }

        var commandPrefix = loadCommandPrefices[type];
        if (commandPrefix) {
            client.send('scpi-quiet', commandPrefix + path + '","' + location + '"', function (response) {
                typeof callback !== 'function' || callback(response);
            });
        }

        if(type == 'rsm'){
            disableExitRecallMeasurement(false);
        }
    }

    function copy(srcpath, srclocation, dstpath, dstlocation, callback) {
        client.send('scpi-quiet', 'mmem:copy "' + srcpath + '","' + srclocation + '","' + dstpath + '","' + dstlocation + '"',
                function(response) {

            typeof callback !== 'function' || callback(response);
        });
    }

    function deleteFile(path, location, callback) {
        client.send('scpi-quiet', 'mmem:del:fil "' + path + '","' + location + '"',
                function(response) {

            typeof callback !== 'function' || callback(response);
        });
    }

    function deleteDir(path, location, callback) {
        client.send('scpi-quiet', 'mmem:del:dir "' + path + '","' + location + '"',
                function(response) {

            typeof callback !== 'function' || callback(response);
        });
    }

    function checkExists(path, location, callbackIfTrue, callbackIfFalse) {
        var i = path.lastIndexOf('/');
        var parentPath = (i === -1 || i === 0) ? '/' : path.slice(0, i);
        var name = (i === -1) ? path : path.slice(i + 1);
        scpi.send_quiet('mmem:cat:dir? "' + parentPath + '","' + location + '"', function(response) {
            if (response.charAt(response.length - 1) === '\n') {
                response = response.slice(0, - 1);
            }

            var i = response.indexOf(",{");
            if (i !== -1 && response.charAt(response.length - 1) === '}' &&
                    response.slice(i + 2, response.length - 1).split("},{").map(function(arg) {

                var fc = arg.split(',');
                if (fc.length === 3) {
                    return fc[0];
                }
            }).filter(function(arg) {
                return arg !== undefined;
            }).indexOf(name) !== -1) {
                typeof callbackIfTrue !== 'function' || callbackIfTrue();
            } else {
                typeof callbackIfFalse !== 'function' || callbackIfFalse();
            }
        });
    }

    var file_overwrite_confirmation = modal_overlay('#file-overwrite-confirmation');
    var file_delete_file_confirmation = modal_overlay('#file-delete-file-confirmation');
    var file_delete_dir_confirmation = modal_overlay('#file-delete-dir-confirmation');

    // a common jstree config creator function
    var createTreeConfig = function(getRootName, selectFirstRoot, conditionalSelect, nodeTypes, fileTypes, updateFreeSpace) {
        var hasRootNode = false;

        function createRootNode() {
            var name = (typeof getRootName === 'function') ? getRootName() : null;
            if (name) {
                var selected;
                if (!hasRootNode) {
                    hasRootNode = true;
                    selected = selectFirstRoot;
                } else{
                    selected = false;
                }

                return { "text": name, "type": "root", "state": {  "opened": true, "loaded": false, "selected": selected } };
            }

            return [];
        };

        function createNode(response, nodeTypes, fileTypes) {
            if (response.charAt(response.length - 1) === '\n') {
                response = response.slice(0, - 1);
            }
            var i = response.indexOf(",{");
            if (i === -1 || response.charAt(response.length - 1) !== '}') {
                return [];
            }

            return response.slice(i + 2, response.length - 1).split("},{").map(function(arg) {
                var fc = arg.split(',');
                if (fc.length === 3) {
                    return (fc[1] === "DIRectory")
                            ? { "text": fc[0], "state": { "loaded": false } }
                            : { "text": fc[0], "type": "file" };
                }
            }).filter(function(arg) {
                return arg !== undefined &&
                    (!nodeTypes || arg.type === undefined || nodeTypes.indexOf(nodeType)) &&
                    (!fileTypes || (arg.type !== "file" || fileTypes.some(function(fileType) {
                        return fileType === arg.text.slice(-fileType.length);
                    })))
            });
        };

        return {
            "core": {
                "multiple": false,
                "force_text": true,
                "data": function(obj, callback) {
                    if (obj.id === "#") {
                        callback.call(this, createRootNode());
                    } else {
                        var path = this.get_path(obj.id, '/');
                        var i = path.indexOf('/');
                        var root = (i === -1) ? obj.text : path.slice(0, i);
                        path = (i === -1) ? '/' : path.slice(i);

                        scpi.send_quiet('mmem:cat:dir? "' + path + '","' + root + '"', function(response) {
                            if (i === -1) {
                                // it should be enough to update free space once
                                // per root dir listing
                                var i1 = response.indexOf(',');
                                var i2 = response.indexOf(',', ++i1);
                                if (i1 > 0 && (typeof updateFreeSpace === 'function')) {
                                    updateFreeSpace(format_bytes((i2 > 0) ? response.slice(i1, i2) : response.slice(i1)));
                                }
                            }
                            callback.call(this, createNode(response,
                                    (typeof nodeTypes === 'function' && nodeTypes(obj)) || nodeTypes,
                                    (typeof fileTypes === 'function' && fileTypes(obj)) || fileTypes));
                        });
                    }
                }
            },
            "types": {
                "#": { "valid_children": ["root"] },
                "root": { "icon": false, "valid_children": ["default", "file"] },
                "default": { "valid_children": ["default","file"] },
                "file": { "icon": "jstree-file", "valid_children": [] }
            },
            "sort": function(obj1, obj2) {
                var sg1 = -1;
                var sg2 = -1;
                var t1 = this.get_type(obj1);
                var t2 = this.get_type(obj2);

                // find the sort groups according to positions in the "types"
                // mapping above
                //
                // "#" and "default" types are special to jstree library, so
                // their positions are always equal to 0 and 1 resp.
                var i = 0;
                for (var k in this.settings.types) {
                    if (k === t1) {
                      sg1 = i;
                        if (sg2 !== -1) {
                            break;
                        }
                    }
                    if (k === t2) {
                        sg2 = i;
                        if (sg1 !== -1) {
                            break;
                        }
                    }
                    ++i;
                }

                // first sort by the sort group
                // then sort by the node text within the same sort group
                return (sg1 === sg2)
                        ? ((this.get_text(obj1).toLowerCase() < this.get_text(obj2).toLowerCase()) ? -1 : 1)
                        : ((sg1 < sg2) ? -1 : 1);
            },
            "conditionalselect": conditionalSelect,
            "plugins": ["sort", "types", "conditionalselect"]
        }
    };

    function setupSaveOnEventConfig(save_on_event_type) {

        var tree_to_return = createTreeConfig(
            function() {
                return $('#save-on-event-location-' + save_on_event_type).val();
            },
            true,
            function(obj) {
                var type = this.get_type(obj);
                if (type === "file") {
                    var name = this.get_text(obj);
                    $('.name-input', '#save-on-event-dialog').val(name);
                    this.deselect_all();
                    this.select_node(this.get_parent(obj));

                    return false;
                }

                return true;
            },
            null,
            function() {
                return [''];
            },
            function(free) {
                $('.free-space', '#save-on-event-dialog').text(free);
            }
        );

        $('.tree', '#save-on-event-dialog').on('changed.jstree', function (e, data) {
            $('.choose-button', '#save-on-event-dialog').attr('disabled', !data || !data.selected || data.selected.length == 0
                    || data.instance.get_type(data.selected[0]) === "file" || !$('.name-input', '#save-on-event-dialog').val());
        });

        $('.name-input', '#save-on-event-dialog').on('input', function(e) {
            var tree = $('.tree', '#save-on-event-dialog').jstree(true);
            var objs = tree.get_selected();
            $('.choose-button', '#save-on-event-dialog').attr('disabled', !objs || objs.length == 0 || tree.get_type(objs[0]) === "file" ||
                    !$(this).val());
        });

        $('.choose-button', '#save-on-event-dialog').on('click', function() {
            var dirname = $('.name-input', '#save-on-event-dialog').val();
            if (dirname) {
                var tree = $('.tree', '#save-on-event-dialog').jstree(true);
                var objs = tree.get_selected();
                if (objs && objs.length > 0) {
                    var obj = objs[0];
                    if (tree.get_type(obj) !== "file") {

                        var path = tree.get_path(obj, '/');
                        var i = path.indexOf('/');
                        path = ((i === -1) ? '/' : path.slice(i) + '/') + dirname;
                        $('#save-on-event-path-' + save_on_event_type).attr('value', path).data('jstree-node-id', obj).data('filename', dirname);
                    }
                }
            }

            fileDialogs["#save-on-event-location-" + save_on_event_type].hide();
        });
	return tree_to_return;
    };


    var limitSaveOnEventTreeConfig = setupSaveOnEventConfig('limit');
    var cmsSaveOnEventTreeConfig =  setupSaveOnEventConfig('cms');

    // file save dialog
    var fileSaveTreeConfig = createTreeConfig(
        function() {
            return $('#file-save-location').val();
        },
        true,
        function(obj) {
            var type = this.get_type(obj);
            if (type === "file") {
                var name = this.get_text(obj);
                $('.name-input', '#file-save-dialog').val(name);
                this.deselect_all();
                this.select_node(this.get_parent(obj));

                return false;
            }

            return true;
        },
        null,
        function() {
            return ['.' + $('#file-save-type').val()];
        },
        function(free) {
            $('.free-space', '#file-save-dialog').text(free);
        }
    );

    $('.tree', '#file-save-dialog').on('changed.jstree', function (e, data) {
        $('.choose-button', '#file-save-dialog').attr('disabled', !data || !data.selected || data.selected.length == 0
                || data.instance.get_type(data.selected[0]) === "file" || !$('.name-input', '#file-save-dialog').val());
    });

    $('.name-input', '#file-save-dialog').on('input', function(e) {
        var tree = $('.tree', '#file-save-dialog').jstree(true);
        var objs = tree.get_selected();
        $('.choose-button', '#file-save-dialog').attr('disabled', !objs || objs.length == 0 || tree.get_type(objs[0]) === "file" ||
                !$(this).val());
    });

    $('.choose-button', '#file-save-dialog').on('click', function() {
        var filename = $('.name-input', '#file-save-dialog').val();
        if (filename) {
            var tree = $('.tree', '#file-save-dialog').jstree(true);
            var objs = tree.get_selected();
            if (objs && objs.length > 0) {
                var obj = objs[0];
                if (tree.get_type(obj) !== "file") {
                    var type = $('#file-save-type').val();
                    var i = filename.indexOf('.');
                    if (i === -1) {
                        filename += '.' + type;
                    }
                    var path = tree.get_path(obj, '/');
                    var i = path.indexOf('/');
                    path = ((i === -1) ? '/' : path.slice(i) + '/') + filename;
                    $('#file-save-path').attr('value', path).data('jstree-node-id', obj).data('filename', filename);
                    $('#file-save-button').attr('disabled', false);
                }
            }
        }

        fileDialogs['#file-save-location'].hide();
    });

    // file recall dialog
    var fileRecallTreeConfig = createTreeConfig(
        function() {
            return $('#file-recall-location').val();
        },
        false,
        function(obj) {
            return this.get_type(obj) === "file";
        },
        null,
        function() {
            var type = $('#file-recall-type').val();
            return (type !== '') ? ['.' + type] : ['.stp', '.rsm', '.lim'];
        },
        function(free) {
            $('.free-space', '#file-recall-dialog').text(free);
        }
    );

    $('.tree', '#file-recall-dialog').on('changed.jstree', function (e, data) {
        $('.choose-button', '#file-recall-dialog').attr('disabled', !data || !data.selected || data.selected.length == 0
                || data.instance.get_type(data.selected[0]) !== "file");
    });

    $('.choose-button', '#file-recall-dialog').on('click', function() {
        var tree = $('.tree', '#file-recall-dialog').jstree(true);
        var objs = tree.get_selected();
        if (objs && objs.length > 0) {
            var obj = objs[0];
            if (tree.get_type(obj) === "file") {
                var path = tree.get_path(obj, '/');
                var i = path.indexOf('/');
                path = (i === -1) ? '/' : path.slice(i);
                $('#file-recall-path').attr('value', path).data('jstree-node-id', obj).data('file', null);
                $('#file-recall-button').attr('disabled', false);
            }
        }

        fileDialogs['#file-recall-location'].hide();
    });

    // file copy from dialog
    var fileCopyFromTreeConfig = createTreeConfig(
        function() {
            return $('#file-copy-from-location').val();
        },
        false,
        function(obj) {
            return this.get_type(obj) === "file";
        },
        null,
        null,
        function(free) {
            $('.free-space', '#file-copy-from-dialog').text(free);
        }
    );

    $('.tree', '#file-copy-from-dialog').on('changed.jstree', function (e, data) {
        $('.choose-button', '#file-copy-from-dialog').attr('disabled', !data || !data.selected || data.selected.length == 0
                || data.instance.get_type(data.selected[0]) !== "file");
    });

    $('.choose-button', '#file-copy-from-dialog').on('click', function() {
        var tree = $('.tree', '#file-copy-from-dialog').jstree(true);
        var objs = tree.get_selected();
        if (objs && objs.length > 0) {
            var obj = objs[0];
            if (tree.get_type(obj) === "file") {
                var path = tree.get_path(obj, '/');
                var i = path.indexOf('/');
                path = (i === -1) ? '/' : path.slice(i);
                $('#file-copy-from-path').attr('value', path).data('jstree-src-node-id', obj);
                updateFileCopyButton();
            }
        }

        fileDialogs['#file-copy-from-location'].hide();
    });

    // file copy to dialog
    var fileCopyToTreeConfig = createTreeConfig(
        function() {
            return $('#file-copy-to-location').val();
        },
        true,
        function(obj) {
            var type = this.get_type(obj);
            if (type === "file") {
                var name = this.get_text(obj);
                $('.name-input', '#file-copy-to-dialog').val(name);
                this.deselect_all();
                this.select_node(this.get_parent(obj));

                return false;
            }

            return true;
        },
        null,
        null,
        function(free) {
            $('.free-space', '#file-copy-to-dialog').text(free);
        }
    );


    $('.tree', '#file-copy-to-dialog').on('changed.jstree', function (e, data) {
        $('.choose-button', '#file-copy-to-dialog').attr('disabled', !data || !data.selected || data.selected.length == 0
                || data.instance.get_type(data.selected[0]) === "file" || !$('.name-input', '#file-copy-to-dialog').val());
    });

    $('.name-input', '#file-copy-to-dialog').on('input', function(e) {
        var tree = $('.tree', '#file-copy-to-dialog').jstree(true);
        var objs = tree.get_selected();
        $('.choose-button', '#file-copy-to-dialog').attr('disabled', !objs || objs.length == 0 ||
                tree.get_type(objs[0]) === "file" || !$(this).val());
    });

    $('.choose-button', '#file-copy-to-dialog').on('click', function() {
        var tree = $('.tree', '#file-copy-to-dialog').jstree(true);
        var objs = tree.get_selected();
        if (objs && objs.length > 0) {
            var obj = objs[0];
            if (tree.get_type(obj) !== "file") {
                var path = tree.get_path(obj, '/');
                var i = path.indexOf('/');
                path = (i === -1) ? '/' : path.slice(i);
                $('#file-copy-to-path').attr('value', path).data('jstree-dst-node-id', obj);
                updateFileCopyButton();
            }
        }

        fileDialogs['#file-copy-to-location'].hide();
    });
    $('.choose-button', '#file-copy-to-dialog').on('click', function() {
        var filename = $('.name-input', '#file-copy-to-dialog').val();
        if (filename) {
            var tree = $('.tree', '#file-copy-to-dialog').jstree(true);
            var objs = tree.get_selected();
            if (objs && objs.length > 0) {
                var obj = objs[0];
                if (tree.get_type(obj) !== "file") {
                    var path = tree.get_path(obj, '/');
                    var i = path.indexOf('/');
                    path = ((i === -1) ? '/' : path.slice(i) + '/') + filename;
                    $('#file-copy-to-path').attr('value', path).data('jstree-dst-node-id', obj).data('dst-filename', filename);
                    updateFileCopyButton();
                }
            }
        }

        fileDialogs['#file-save-location'].hide();
    });

    // file delete dialog
    var fileDeleteTreeConfig = createTreeConfig(
        function() {
            return $('#file-delete-location').val();
        },
        false,
        function(obj) {
            var type = this.get_type(obj);

            return type === "default" || type === "file";
        },
        null,
        null,
        function(free) {
            $('.free-space', '#file-delete-dialog').text(free);
        }
    );

    $('.tree', '#file-delete-dialog').on('changed.jstree', function (e, data) {
        $('.choose-button', '#file-delete-dialog').attr('disabled', !data || !data.selected || data.selected.length == 0 ||
                (data.instance.get_type(data.selected[0]) !== "default" &&
                data.instance.get_type(data.selected[0]) !== "file"))

    });

    $('.choose-button', '#file-delete-dialog').on('click', function() {
        var tree = $('.tree', '#file-delete-dialog').jstree(true);
        var objs = tree.get_selected();
        if (objs && objs.length > 0) {
            var obj = objs[0];
            var path = tree.get_path(obj, '/');
            var i = path.indexOf('/');
            path = (i === -1) ? '/' : path.slice(i);
            $('#file-delete-path').attr('value', path).data('jstree-node-id', obj).data('isdir',
                    tree.get_type(obj) === "default");
            $('#file-delete-button').attr('disabled', false);
        }
        fileDialogs['#file-delete-location'].hide();
    });

    // file overwrite confirmation
    $('.confirm', '#file-overwrite-confirmation').on('click', function() {
        var $el = $('#file-overwrite-confirmation');

        switch ($el.data('action')) {
        case 'save':
            save($el.data('type'), $el.data('path'), $el.data('location'));
            break;
        case 'upload':
            upload($el.data('file'), $el.data('path'), $el.data('location'));
            break;
        case 'copy':
            copy($el.data('srcpath'), $el.data('srclocation'), $el.data('dstpath'), $el.data('dstlocation'));
            break;
        }

        file_overwrite_confirmation.hide();
    });

    // file delete file confirmation
    $('.confirm', '#file-delete-file-confirmation').on('click', function() {
        var $el = $('#file-delete-file-confirmation');
        deleteFile($el.data('path'), $el.data('location'));

        file_delete_file_confirmation.hide();
    });

    // file delete dir confirmation
    $('.confirm', '#file-delete-dir-confirmation').on('click', function() {
        var $el = $('#file-delete-dir-confirmation');
        deleteDir($el.data('path'), $el.data('location'));

        file_delete_dir_confirmation.hide();
    });

    // save settings group

    $('#file-save-type').on('change', function(e) {
        $('#file-save-path').attr('value', null).data('jstree-node-id', null).data('filename', null);
        updateFileSaveButton();
    });

    $('#file-save-location').on('change', function(e) {
        $('#file-save-path').attr('value', null).data('jstree-node-id', null).data('filename', null);
        var file_save_location = $('#file-save-location option:selected').text();
        prevSelectedLocation["#file-save-location"] = file_save_location;
        updateFileSavePath();
        updateFileSaveButton();
    });

    $('#save-on-event-location-limit').on('change', function(e) {

        var new_location = $('#save-on-event-location-limit option:selected').text();
        prevSelectedLocation["#save-on-event-location-limit"] = new_location;

    });

    $('#save-on-event-location-cms').on('change', function(e) {

        var new_location = $('#save-on-event-location-cms option:selected').text();
        prevSelectedLocation["#save-on-event-location-cms"] = new_location;

    });


    function updateListofDevices(){
        scpi.send_quiet("mmem:cat:msus?", function(response) {
            var available_msuses = ((response.charAt(response.length - 1) !== '\n') ? response : response.slice(0, - 1)).split(',');
            update_msus_combo(available_msuses);
        })
    };

    $('#file-save-choose-button').on('click', function() {
        if ($('#file-save-location').prop('selectedIndex') !== 0) {
            var file_dialog_tree = $('.tree', '#file-save-dialog');
            var tree = file_dialog_tree.jstree(true);

            if (tree)
            {
                var obj = $('#file-save-path').data('jstree-node-id');
                tree.deselect_all();
                if (obj)
                {
                    tree.select_node(obj);
                }

                else
                {
                    var roots = tree.get_node("#").children;
                    if (roots && roots.length > 0)
                    {
                        tree.select_node(roots[0]);
                    }
                }
                tree.refresh();
                $('.name-input', '#file-save-dialog').val($('#file-save-path').data('filename'));
            }
            else
            {
                file_dialog_tree.jstree(fileSaveTreeConfig);
            }

            fileDialogs["#file-save-location"].show();
        }
    });

    function processSaveOnEventClick(save_on_event_type, treeConfig) {
        var file_dialog_tree = $('.tree', '#save-on-event-dialog');
        var tree = file_dialog_tree.jstree(true);

        if (tree)
        {
            var obj = $('#file-save-path').data('jstree-node-id');
            tree.deselect_all();
            if (obj)
            {
                tree.select_node(obj);
            }

            else
            {
                var roots = tree.get_node("#").children;
                if (roots && roots.length > 0)
                {
                    tree.select_node(roots[0]);
                }
            }
            tree.refresh();
            $('.name-input', '#save-on-event-dialog').val($('#save-on-event-path-' + save_on_event_type).data('filename'));
        }
        else
        {
            file_dialog_tree.jstree(treeConfig);
        }

        fileDialogs["#save-on-event-location-" + save_on_event_type].show();
    };

    $('#save-on-event-choose-button-limit').on('click', function() {
                processSaveOnEventClick('limit', limitSaveOnEventTreeConfig);
            });

    $('#save-on-event-choose-button-cms').on('click', function() {
                processSaveOnEventClick('cms', cmsSaveOnEventTreeConfig);
            });

    function processSaveOnEventCommit(save_on_event_type) {

        var path = $('#save-on-event-path-' + save_on_event_type).attr('value');
        var location = $('#save-on-event-location-' + save_on_event_type).val();

        client.send('scpi-quiet', 'mmem:msis "' + location + '"', function (response) {
            typeof callback !== 'function'|| callback(response);
        });

        client.send('scpi-quiet', 'mmem:cre:dir "' + path + '","' + location + '"', function (response) {
            typeof callback !== 'function'|| callback(response);
        });

        client.send('scpi-quiet', 'mmem:cdir "' + path + '"', function (response) {
            typeof callback !== 'function'|| callback(response);
        });


        updateSaveOnEventCommitButton();
    };

    $('#save-on-event-commit-button-limit').on('click', function() {
        processSaveOnEventCommit('limit');
    });

    $('#save-on-event-commit-button-cms').on('click', function() {
        processSaveOnEventCommit('cms');
    });

    $('#file-save-button').on('click', function() {
        if ($('#file-save-location').prop('selectedIndex') === 0) {
            if (firstLocation) {
                var type = $('#file-save-type').val();
                var path = "/.tmp/save." + type;
                save(type, path, firstLocation, function(response) {
                    if (!response.errors) {
                        download(path, firstLocation, "file." + type, function() {
                            deleteFile(path, firstLocation);
                        });
                    } else {
                        deleteFile(path, firstLocation);
                    }
                });
            }
        } else {
            var type = $('#file-save-type').val();
            var path = $('#file-save-path').attr('value');
            var location = $('#file-save-location').val();
            if (path && location) {
                checkExists(path, location, function() {
                    $('#file-overwrite-confirmation').data('action', 'save')
                            .data('type', type).data('path', path).data('location', location);
                    $('.file-name', '#file-overwrite-confirmation').text(path);
                    $('.location', '#file-overwrite-confirmation').text(location);

                   file_overwrite_confirmation.show();
                }, function() {
                           save(type, path, location);
                        });
               }

           $('#file-save-path').attr('value', null).data('jstree-node-id', null).data('filename', null);
           updateFileSaveButton();
        }
    });

    function updateFileSavePath() {
        var disabled = $('#file-save-location').prop('selectedIndex') === 0;
        $('#file-save-path-label').attr('disabled', disabled);
        $('#file-save-path').attr('disabled', disabled);
        $('#file-save-choose-button').attr('disabled', disabled);
    }

    function updateSaveOnEventCommitButton() {
        $('#save-on-event-commit-button').attr('disabled',
                $('#save-on-event-location').prop('selectedIndex') !== 0 &&
                        (!$('#save-on-event-path').data('jstree-node-id') || !$('#save-on-event-path').data('filename')));
    }

    function updateFileSaveButton() {
        $('#file-save-button').attr('disabled',
                $('#file-save-location').prop('selectedIndex') !== 0 &&
                        (!$('#file-save-path').data('jstree-node-id') || !$('#file-save-path').data('filename')));
    }

    // recall settings group
    $('#file-recall-type').on('change', function(e) {
        $('#file-recall-path').attr('value', null).data('jstree-node-id', null).data('file', null);
        updateFileRecallButton();
    });

    $('#file-recall-location').on('change', function(e) {
        $('#file-recall-path').attr('value', null).data('jstree-node-id', null).data('file', null);
        var file_recall_location = $('#file-recall-location option:selected').text();
        prevSelectedLocation["#file-recall-location"] = file_recall_location;
        updateFileRecallButton();
    });

    $('#file-recall-choose-input').on('change', function(e) {
        var file = e.target.files[0];
        $('#file-recall-path').attr('value', (file) ? $(this).val() : null).data('file', file);
        updateFileRecallButton();
    });

    $('#file-recall-choose-button').on('click', function() {
        if ($('#file-recall-location').prop('selectedIndex') === 0) {
            var type = $('#file-recall-type').val();
            $('#file-recall-choose-input').attr('accept', (type) ? '.' + type : '.stp,.rsm,.lim').click();
        } else {
            var file_dialog_tree = $('.tree', '#file-recall-dialog');
            var tree = file_dialog_tree.jstree(true);
            if (tree)
            {
                var obj = $('#file-recall-path').data('jstree-node-id');
                tree.deselect_all();
                if (obj)
                {
                    tree.select_node(obj);
                }
                tree.refresh();
            }
            else
            {
                file_dialog_tree.jstree(fileRecallTreeConfig);
            }

            fileDialogs["#file-recall-location"].show();
         }
    });

    function disableExitRecallMeasurement(state){
        $('#file-recall-load-previous-button').attr('disabled',state);
        $('#file-recall-keep-current-button').attr('disabled',state);
    }

    function exitRecallMeasurement(mode){
        client.send('scpi-quiet','mmem:load:trac:exit '+ mode);
        disableExitRecallMeasurement(true);
    }

    $('#file-recall-load-previous-button').on('click',function(){
        exitRecallMeasurement('load');
    });

    $('#file-recall-keep-current-button').on('click',function(){
        exitRecallMeasurement('keep');
    });

    $('#file-recall-button').on('click', function() {
        if ($('#file-recall-location').prop('selectedIndex') === 0) {
            var file = $('#file-recall-path').data('file');
            if (file && firstLocation) {
                var path = "/.tmp/" + file.name;
                upload(file, path, firstLocation, function(response) {
                    if (!response.errors) {
                        var type = $('#file-recall-type').val();
                        recall(type, path, firstLocation, function() {
                            deleteFile(path, firstLocation);
                        });
                    }
                });
            }
        } else {
            var type = $('#file-recall-type').val();
            var path = $('#file-recall-path').attr('value');
            var location = $('#file-recall-location').val();
            if (path && location) {
                recall(type,path,location);
            }

            $('#file-recall-path').attr('value', null).data('jstree-node-id', null).data('file', null);
            updateFileRecallButton();
        }
    });

    // copy settings group
    $('#file-copy-from-location').on('change', function(e) {
        $('#file-copy-from-path').attr('value', null).data('jstree-src-node-id', null).data('src-file', null);
        var file_copy_from_location = $('#file-copy-from-location option:selected').text();
        prevSelectedLocation["#file-copy-from-location"] = file_copy_from_location;
        updateFileCopyFromPath();
        updateFileCopyToPath();
        updateFileCopyButton();
    });

    $('#file-copy-to-location').on('change', function(e) {
        $('#file-copy-to-path').attr('value', null).data('jstree-dst-node-id', null).data('dst-filename', null);
        var file_copy_to_location = $('#file-copy-to-location option:selected').text();
        prevSelectedLocation["#file-copy-to-location"] = file_copy_to_location;
        updateFileCopyFromPath();
        updateFileCopyToPath();
        updateFileCopyButton();
    });

    $('#file-copy-from-choose-input').on('change', function(e) {
        var file = e.target.files[0];
        $('#file-copy-from-path').attr('value', (file) ? $(this).val() : null).data('src-file', file);
        updateFileCopyButton();
    });

    $('#file-copy-from-choose-button').on('click', function() {
        if ($('#file-copy-from-location').prop('selectedIndex') === 0) {
            $('#file-copy-from-choose-input').click();
        } else {
            var file_dialog_tree = $('.tree', '#file-copy-from-dialog');
            var tree = file_dialog_tree.jstree(true);
            if (tree)
            {
                var obj = $('#file-copy-from-path').data('jstree-src-node-id');
                tree.deselect_all();
                if (obj)
                {
                    tree.select_node(obj);
                }
                tree.refresh();
            }
            else
            {
                file_dialog_tree.jstree(fileCopyFromTreeConfig);
            }
            fileDialogs["#file-copy-from-location"].show();
        }
    });

    $('#file-copy-to-choose-button').on('click', function() {
        if ($('#file-copy-to-location').prop('selectedIndex') !== 0) {
            var file_dialog_tree = $('.tree', '#file-copy-to-dialog');
            var tree = file_dialog_tree.jstree(true);
            if (tree)
            {
                var obj = $('#file-copy-to-path').data('jstree-dst-node-id');
                tree.deselect_all();
                if (obj)
                {
                    tree.select_node(obj);
                }
                else
                {
                    var roots = tree.get_node("#").children;
                    if (roots && roots.length > 0)
                    {
                        tree.select_node(roots[0]);
                    }
                }
                tree.refresh();
            }
            else
            {
                file_dialog_tree.jstree(fileCopyToTreeConfig);
            }

            var name = $('#file-copy-to-path').data('dst-filename');
            if (!name)
            {
                if ($('#file-copy-from-location').prop('selectedIndex') === 0 && $('#file-copy-from-path').data('src-file')) {
                    name = $('#file-copy-from-path').data('src-file').name;
                }
                else
                {
                    var path = $('#file-copy-from-path').attr('value');
                    if (path)
                    {
                        var i = path.lastIndexOf('/');
                        name = (i === -1) ? path : path.slice(i + 1);
                    }
                }
            }
            $('.name-input', '#file-copy-to-dialog').val(name);

            fileDialogs["#file-copy-to-location"].show();
         }
    });

    $('#file-copy-button').on('click', function() {

       if ($('#file-copy-from-location').prop('selectedIndex') === 0) {
            var file = $('#file-copy-from-path').data('src-file');
            var path = $('#file-copy-to-path').attr('value');
            var location = $('#file-copy-to-location').val();
            if (file && path && location) {
                checkExists(path, location, function() {
                    $('#file-overwrite-confirmation').data('action', 'upload')
                            .data('file', file).data('path', path).data('location', location);
                    $('.file-name', '#file-overwrite-confirmation').text(path);
                    $('.location', '#file-overwrite-confirmation').text(location);

                   file_overwrite_confirmation.show();
                }, function() {
                    upload(file, path, location);
                });
            }
        } else if ($('#file-copy-to-location').prop('selectedIndex') === 0) {
            var path = $('#file-copy-from-path').attr('value');
            var location = $('#file-copy-from-location').val();
            if (path && location) {
                var i = path.lastIndexOf('/');
                download(path, location, (i === -1) ? path : path.slice(i + 1));
            }
        } else {
            var srcpath = $('#file-copy-from-path').attr('value');
            var srclocation = $('#file-copy-from-location').val();
            var dstpath = $('#file-copy-to-path').attr('value');
            var dstlocation = $('#file-copy-to-location').val();
            if (srcpath && srclocation && dstpath && dstlocation) {
                checkExists(dstpath, dstlocation, function() {
                    $('#file-overwrite-confirmation').data('action', 'copy')
                            .data('srcpath', srcpath).data('srclocation', srclocation)
                            .data('dstpath', dstpath).data('dstlocation', dstlocation);
                    $('.file-name', '#file-overwrite-confirmation').text(dstpath);
                    $('.location', '#file-overwrite-confirmation').text(dstlocation);

                   file_overwrite_confirmation.show();
                }, function() {
                    copy(srcpath, srclocation, dstpath, dstlocation);
                });
            }
        }

        $('#file-copy-from-path').attr('value', null).data('jstree-src-node-id', null).data('src-file', null);
        $('#file-copy-to-path').attr('value', null).data('jstree-dst-node-id', null).data('dst-filename', null);
        updateFileCopyButton();
    });

    // delete settings group
    $('#file-delete-location').on('change', function(e) {
        $('#file-delete-path').attr('value', null).data('jstree-node-id', null).data('isdir', null);
        var file_delete_location = $('#file-delete-location option:selected').text();
        prevSelectedLocation["#file-delete-location"] = file_delete_location;
        updateFileDeleteButton();
    });

    $('#file-delete-choose-button').on('click', function() {
        var file_dialog_tree = $('.tree', '#file-delete-dialog');
        var tree = file_dialog_tree.jstree(true);
        if (tree)
        {
            var obj = $('#file-delete-path').data('jstree-node-id');
            tree.deselect_all();
            if (obj)
            {
                tree.select_node(obj);
            }
            tree.refresh();
        }
        else
        {
            file_dialog_tree.jstree(fileDeleteTreeConfig);
        }

        fileDialogs["#file-delete-location"].show();
    });

    $('#file-delete-button').on('click', function() {

        var path = $('#file-delete-path').attr('value');
        var location = $('#file-delete-location').val();
        if (path && location)
        {
            if ($('#file-delete-path').data('isdir'))
            {
                $('#file-delete-dir-confirmation').data('path', path).data('location', location);
                $('.dir-name', '#file-delete-dir-confirmation').text(path);
                $('.location', '#file-delete-dir-confirmation').text(location);
                file_delete_dir_confirmation.show();
            }
            else
            {
                $('#file-delete-file-confirmation').data('path', path).data('location', location);
                $('.file-name', '#file-delete-file-confirmation').text(path);
                $('.location', '#file-delete-file-confirmation').text(location);
                file_delete_file_confirmation.show();
            }
        }

        $('#file-delete-path').attr('value', null).data('jstree-node-id', null).data('isdir', null);
        updateFileDeleteButton();
    });
}

var active_tab_name;
var prevSelectedLocation = {};
var locations = ["#file-save-location",
                 "#file-recall-location",
                 "#file-copy-from-location",
                 "#file-copy-to-location",
                 "#file-delete-location",
                 "#save-on-event-location-cms",
                 "#save-on-event-location-limit"];

var notification_board = modal_overlay('#notification');

// this functionality can be used to display custom notifications over the gps data
// color : specify color for messages of different purpose
// timed : defaults to none in the case of which the user has to dimiss by clicking on it
//       : or specify time in seconds after which the notificaiton disappears
// NOTE: this functionality currently supports messages of smaller size only
function display_notification(message, color, time = 0)
{
    jQuery("#notification-text").text(message +  "[x]");
    $('#notification-text').css('color', color);

    notification_board.show();

    if (time != 0)
    {
        setTimeout(function() {
            notification_board.hide();
        },time*1000);
    }
}

function update_msus_combo(available_msuses)
{
    firstLocation = "";

    ['#file-save-location', '#file-recall-location', '#file-copy-from-location', '#file-copy-to-location',
     '#file-delete-location', '#save-on-event-location-limit', '#save-on-event-location-cms'].forEach(function(selector) {

        $(selector).children('option:gt(0)').remove();

        available_msuses.forEach(function(value, i) {
            if (i === 0) {
                firstLocation = available_msuses[0];
            }

            if ($(selector + " option[value='" + value + "']").length === 0)
                $(selector).append($('<option></option>').attr('value', value).text(value));
        });


     });

    update_msus_combo_helper();
}

function update_msus_combo_helper()
{
    // workaround to disable UI elements, because they are enabled
    // automaticaly in some other onconnect hook
    setTimeout(function() {
    updateFileSavePath();
    updateFileSaveButton();
    updateFileRecallButton();
    updateFileCopyFromPath();
    updateFileCopyToPath();
    updateFileCopyButton();
    updateFileDeleteButton();
    }, 200);
}

function updateFileSavePath() {
    var disabled = $('#file-save-location').prop('selectedIndex') === 0;
    $('#file-save-path-label').attr('disabled', disabled);
    $('#file-save-path').attr('disabled', disabled);
    $('#file-save-choose-button').attr('disabled', disabled);
}

function updateFileSaveButton() {
    $('#file-save-button').attr('disabled',
            $('#file-save-location').prop('selectedIndex') !== 0 &&
                    (!$('#file-save-path').data('jstree-node-id') || !$('#file-save-path').data('filename')));
}

function updateFileRecallButton() {
    $('#file-recall-button').attr('disabled',
            !$('#file-recall-path').data('jstree-node-id') && !$('#file-recall-path').data('file'));
}

function updateFileCopyFromPath() {
    var disabled = $('#file-copy-from-location').prop('selectedIndex') === 0 &&
            $('#file-copy-to-location').prop('selectedIndex') === 0;
    $('#file-copy-from-path-label').attr('disabled', disabled);
    $('#file-copy-from-path').attr('disabled', disabled);
    $('#file-copy-from-choose-button').attr('disabled', disabled);
}

function updateFileCopyToPath() {
    var disabled = $('#file-copy-to-location').prop('selectedIndex') === 0;
    $('#file-copy-to-path-label').attr('disabled', disabled);
    $('#file-copy-to-path').attr('disabled', disabled);
    $('#file-copy-to-choose-button').attr('disabled', disabled);
}

function updateFileCopyButton() {
    var srci = $('#file-copy-from-location').prop('selectedIndex');
    var dsti = $('#file-copy-to-location').prop('selectedIndex');
    var disabled;
    if (srci === 0 && dsti === 0) {
        disabled = true;
    } else if (srci === 0 && dsti !== 0) {
        disabled = !$('#file-copy-from-path').data('src-file') || !$('#file-copy-to-path').data('jstree-dst-node-id')
                            || !$('#file-copy-to-path').data('dst-filename');
    } else if (srci !== 0 && dsti === 0) {
        disabled = !$('#file-copy-from-path').data('jstree-src-node-id');
    } else if (srci !== 0 && dsti !== 0) {
        if (!$('#file-copy-from-path').data('jstree-src-node-id') || !$('#file-copy-to-path').data('jstree-dst-node-id')
                            || !$('#file-copy-to-path').data('dst-filename')) {

            disabled = true;
        } else if (srci === dsti) {
            var srctree = $('.tree', '#file-copy-from-dialog').jstree(true);
            if (!srctree) {
                disabled = true;
            } else {
                var srcpath = srctree.get_path($('#file-copy-from-path').data('jstree-src-node-id'), '/');
                var i = srcpath.indexOf('/');
                srcpath = (i === -1) ? '/' : srcpath.slice(i);

                disabled = (srcpath === $('#file-copy-to-path').attr('value'));
            }
        } else {
            disabled == false;
        }
    }

    $('#file-copy-button').attr('disabled', disabled);
}

function updateFileDeleteButton() {
    $('#file-delete-button').attr('disabled', !$('#file-delete-path').data('jstree-node-id'));
}

function storage_device_changes(available_msuses)
{
    // any changes to storage devices changes will invoke this function
    // this function will lookup if the previously selected #file-xxxxx-location
    // ..is still present in the new list
    // if yes then re-selects the previously selected index
    // else issues a warning notification,

    update_msus_combo(available_msuses);

    var locationsLength = locations.length;

    for(var index = 0; index < locationsLength; index++)
    {
        var prevSelection = prevSelectedLocation[locations[index]];
        if (prevSelection && prevSelection != "default")
        {
            if (available_msuses.indexOf(prevSelection) != -1)
            {
                $(locations[index]).val(prevSelection);
            }

            else
            {
                prevSelectedLocation[locations[index]] = "default";

                var message = "WARNING " + prevSelection + " is unplugged";

                display_notification(message, 'red');

                if (active_tab_name === "file")
                {
                    fileDialogs[locations[index]].hide();
                }
            }
        }
    }
}
