'use strict';

function TraceData(client) {
    var raw_trace_data = [];
    var ref_level;
    var units;
    var amplitude_span = 100;
    var on_data = [];
    var stale_mask = [];
    var status_mask = [];
    var trace_pairs = [];
    var sweep_id;
    var paused = false;
    var speed_ms = -1;

    var frequency = {
        start: null,
        stop: null,
        center: null,
        span: null,
        to_relative: function (f) {
            return (f - this.start) / this.span;
        },
        from_relative: function (relative) {
            return this.start + this.span * relative;
        }
    };

    client.onconnect(function () {
        ref_level = undefined;
        units = "DBM";
    });

    var poll_handle = client.poll('trace-data', function handle_trace_data(val) {
        if (!val) return;
        if (typeof ref_level === 'undefined') return;

        var points = [];
        var num_points = val.count;
        var start_index = val.start;
        sweep_id = val.sweep_id;
        speed_ms = val.speed_ms;

        for (var i = 0, chunks = val.data.length / 9; i < chunks; i++) {
            var sign = val.data[i*9] == '-' ? -1 : 1;
            var hex_bytes = val.data.substr(i * 9 + 1, 8);
            var int_val = sign * Math.abs(parseInt(hex_bytes, 16));
            raw_trace_data[i + start_index] = (int_val / 1000);
        }

        raw_trace_data = raw_trace_data.slice(0, num_points);
        stale_mask = stale_mask.slice(0, num_points);
        status_mask = status_mask.slice(0, num_points);

        for (var i = 0, chunks = val.stale.length; i < chunks; i++) {
            stale_mask[i + start_index] = parseInt(val.stale[i]);
        }

        for (var i = 0, chunks = val.status.length / 8; i < chunks; i++) {
            var bytes = val.status.substr(i * 8, 8);
            status_mask[i + start_index] = parseInt(bytes, 16);
        }

        trace_pairs = raw_trace_data.map(function (amp, i) {
            var relative_freq = i / num_points;
            return {
                amp: amp,
                freq: frequency.from_relative(relative_freq)
            };
        });

        on_data.map(call);
    });

    return {
        set_ref_level: function (x) {
            ref_level = x;
        },
        ref_level: function () { return ref_level; },
        set_units: function (x) { 
            units = x; 
        },
        units: function () { return units; },
        set_amplitude_span: function (x) {
            amplitude_span = x;
        },
        get_ref_labels: function (count) {
            var ret = [];
            for (var i = 0; i < count; i++) {
                // TODO: this dB/div calculation will not work for linear units
                var amp = ref_level - i * amplitude_span / count;
                var label;
                if (isNaN(amp))
                    label = '';
                else if (i === 0 || i === count - 1)
                    label = format_amplitude(amp, units);
                else
                    label = format_amplitude(amp, units).split(' ')[0];
                ret.push(label);
            }
            return ret;
        },
        data: function () {
            return raw_trace_data;
        },
        pairs: function () {
            return trace_pairs;
        },
        denormalize: function (ratio) {
            // inverse of normalize, maybe better named from_relative
            return (ratio * amplitude_span) + ref_level - amplitude_span;
        },
        normalize: function (ampl) {
            return (ampl - ref_level + amplitude_span) / amplitude_span;
        },
        normalized_trace_data: function () {
            var that = this;
            return raw_trace_data.map(function (ampl) {
                return that.normalize(ampl);
            });
        },
        on_data: function on_trace_data(cb) {
            on_data.push(cb);
        },
        stale_mask: function () {
            return stale_mask;
        },
        status_mask: function () {
            return status_mask;
        },
        frequency: frequency,
        amplitude_at_freq: function(freq) {
            if (isNaN(freq) || freq < frequency.start || freq > frequency.stop)
                return {};
            if (raw_trace_data) {
                var relative_freq = frequency.to_relative(freq)
                var index = Math.round(relative_freq * (raw_trace_data.length - 1));
                if (!isNaN(raw_trace_data[index])) {
                    return {
                        freq: frequency.from_relative(index / (raw_trace_data.length - 1)),
                        amp: raw_trace_data[index],
                    }
                }
            }
            return {};
        },
        id: function () {
            return sweep_id;
        },
        pause: poll_handle.pause,
        unpause: poll_handle.resume,
        set_trace_refresh_parameters : function(params){
            poll_handle.update_poll_rate(params);
        },

        pretty_sweep_speed: function () {
            if (speed_ms < 0 || array_any(stale_mask)) {
                return 'N/A';
            }
            else if (parseInt(speed_ms) === 0) {
                return '< 1 ms';
            }
            else {
                return format_microseconds(speed_ms * 1e3);
            }
        }
    }
};
