'use strict';

function json_tcp_client(websocket_factories, status) {
    var sock = null;
    var buffer = '';
    var room_callbacks = {};
    var pending_timeouts = {};
    var ack_callbacks = {};
    var onconnect_callbacks = [];
    var ondisconnect_callbacks = [];
    var methods = {};

    status = status || function () {};

    function call(f) { return f(); }

    var next_websocket_factory = call(function () {
        var index = 0;
        return function () {
            index = (index + 1) % websocket_factories.length;
            return websocket_factories[index]
        }
    });

    function try_json_parse(message_buffer) {
        // Chrome is quirky and does not optimize any function that contains a
        // try/catch block. The only way to use JSON.parse() reliably is with
        // a try/catch block. So, this function exists to that Chrome will
        // optimize the rest of consume_message().
        try {
            return JSON.parse(message_buffer);
        }
        catch (e) {
            status('Invalid JSON: "' + message_buffer + '"');
            return null;
        }
    }

    function consume_message(message_buffer) {
        // This function accepts one single line that is a complete JSON
        // object with no trailing newline

        var response = try_json_parse(message_buffer);
        if (!response) {
            return;
        }

        if (response.ack) {
            var callback = ack_callbacks[response.ack];
            delete ack_callbacks[response.ack];
            if (typeof callback === 'function') {
                callback(response);
            }
        }

        if (response.error) {
            console.warn(response);
        }

        //trim out the \n from response
        if(response && response.value && response.value.hasOwnProperty('response'))
        {
            response.value.response = response.value.response.trim();
        }

        var empty_ack = response.ack && !response.value;

        if (!response.error && !empty_ack && room_callbacks[response.type]) {
            room_callbacks[response.type].map(function (callback) {
                if (typeof callback === 'function') {
                    var t = setTimeout(function () {
                        console.error('Bad callback for ' + response.type);
                        delete room_callbacks[response.type];
                        methods.leave(response.type);
                    }, 100);

                    callback(response.value);
                    clearTimeout(t);
                }
            });
        }
    }

    // Polyfill for IE. This is only a polyfill for *ascii* so it's a
    // oneliner. It is *not* some universal TextDecoder polyfill that should
    // be abstracted out into a library.
    var ascii_decoder;
    if (typeof TextDecoder !== 'undefined') {
        ascii_decoder = new TextDecoder('ascii');
    }
    else {
        ascii_decoder = {
            decode: function (buf) {
                return String.fromCharCode.apply(null, new Uint8Array(buf))
            }
        };
    }

    function onmessage(event) {
        if (typeof event.data === 'string') {
            buffer += event.data;
        }
        else {
            // Decode the binary response. The type of the returned data
            // depends on the server, apparently. Specifying binary protocol
            // on lighttpd seems to have no effect, and the event.data is
            // returned as a plain string. Specifying binary protocol on
            // websockify is required, and the returned event.data is an
            // ArrayBuffer.
            buffer += ascii_decoder.decode(event.data);
        }
        // Consume all messages one line at a time
        while (true) {
            // The first character must be an open curly brace
            buffer = buffer.replace(/^[^{]+/, '');

            var newline_pos = buffer.indexOf('\n');
            if (newline_pos < 0) {
                return;
            }

            // Split the buffer on the newline and consume the first part
            consume_message(buffer.substr(0, newline_pos));
            buffer = buffer.slice(newline_pos);
        }
    }

    function onopen() {
        onconnect_callbacks.map(call);
        status('Connected');
    }

    function terminate() {
        if (sock) {
            sock.onmessage = function () {};
            sock.onerror = function () {};
            sock.onopen = function () {};
            sock.onclose = function () {};
            sock.close();
            sock = null;
        }

        ack_callbacks = {};
        room_callbacks = {};
        for (var timeout_id in pending_timeouts) {
            clearTimeout(timeout_id);
        }
        pending_timeouts = {};
        buffer = '';

        ondisconnect_callbacks.map(function (cb) {
            cb();
        });
    }

    function error_handler(msg) {
        return function () {
            status(msg);
            terminate();
        }
    }

    var last_connect = 0;
    function check_connection() {
        if (methods.connected()) {
            return
        }
        else if (sock && sock.readyState == WebSocket.CONNECTING) {
            return;
        }
        else if (sock) {
            terminate();
        }
        else {
            if (Date.now() < last_connect + 1000) {
                return;
            }
            last_connect = Date.now();

            // Firefox and Chrome work with lighttpd+mod_websocket when no
            // protocol is specified. Only Firefox works with lighttpd and
            // protocol 'base64'.
            // I think specifying the protocol is needed with websockify

            // Chrome does not work with lighttpd+mod_websocket if protocols =
            // 'binary' is specified.

            var websocket_factory = next_websocket_factory();
            sock = websocket_factory();
            sock.binaryType = 'arraybuffer';
            sock.onmessage = onmessage;
            sock.onopen = onopen;
            sock.onerror = error_handler('Connection error');
            sock.onclose = error_handler('Connection closed');
        }
    }

    setInterval(check_connection, 100);

    var ack_id = call(function () {
        var id = 0;
        return function () {
            id += 1;
            return id;
        }
    });

    function poll(type, max_in_flight, callback) {
        if (typeof max_in_flight === 'function') {
            callback = max_in_flight;
            max_in_flight = 40;
        }

        var in_flight;
        var interval;
        var last_rx;
        var paused = false;
        var refresh_parameters= {
            manual_refresh: null,
            refresh_rate_in_ms:0
        };

        function update_refresh_rate_parameters(params){
            refresh_parameters = params;
            poll_resume();
        }

        function poll_pause() {
            paused = true;
        }

        function poll_resume() {
            in_flight = 0;

            if(Boolean(refresh_parameters.manual_refresh)){
                interval = refresh_parameters.refresh_rate_in_ms;
            }else{
                interval = 100;
            }

            last_rx = Date.now();
            if (paused) {
                paused = false;
                work();
            }
        }

        methods.onconnect(function () {
            in_flight = 0;
            interval = 100;
            last_rx = Date.now();
        });

        function work() {
            if (methods.connected()) {
                if (in_flight < max_in_flight || Boolean(refresh_parameters.manual_refresh)) {
                    methods.send(type, function () {
                        in_flight--;
                        last_rx = Date.now();
                        callback.apply(null, arguments);
                    });

                    in_flight++;
                }
            }

            if (!paused){
                setTimeout(work, Math.max(16, interval));
            }
        }

        setInterval(function poll_throttle() {
            if (!methods.connected()) {
                return;
            }

            if(Boolean(refresh_parameters.manual_refresh)){
                //no need to calculate interval if manual refresh rate is set by user
                return;
            }

            var rx_time = Date.now() - last_rx;
            var flight_ratio = in_flight / max_in_flight;

            if (rx_time > 1.2 * interval || flight_ratio > 0.6) {
                interval = Math.min(1000, interval * 1.02);
            }
            else if (rx_time < 0.8 * interval || flight_ratio < 0.4) {
                interval = Math.max(10, interval * 0.98);
            }
        }, 100);

        work();

        return {
            pause: poll_pause,
            resume: poll_resume,
            update_poll_rate: update_refresh_rate_parameters,
        };
    }

    methods = {
        connected: function () {
            return sock && sock.readyState == WebSocket.OPEN;
        },
        send: function(message_type, value, onsuccess) {
            if (typeof value === 'function' && typeof onsuccess === 'undefined') {
                onsuccess = value;
                value = undefined;
            }

            if (typeof message_type === 'undefined') {
                throw 'Cannot send a message without a type';
            }
            else if (methods.connected()) {
                var payload = { type: message_type, };

                if (typeof value !== 'undefined')
                    payload.value = value;

                if (typeof onsuccess === 'function') {
                    payload.ack = ack_id();

                    var timeout_func = function () {
                        console.error('Timed out sending ' +
                                      JSON.stringify(payload));
                    }
                    var timeout_id = setTimeout(timeout_func, 15000);
                    pending_timeouts[timeout_id] = timeout_id;

                    ack_callbacks[payload.ack] = function (response) {
                        clearTimeout(timeout_id);
                        delete pending_timeouts[timeout_id];

                        if ( message_type === 'scpi' ) onerror = onsuccess;

                        if (response.error) {

                            if (typeof onerror === 'function') {
                                onerror(response.value);
                            }
                            else {
                                console.error('Error: ' +
                                    JSON.stringify(payload) +
                                    ' -> ' +
                                    JSON.stringify(response));
                            }
                        }
                        else {
                            onsuccess(response.value);
                        }
                    };
                }

                var msg = JSON.stringify(payload) + '\n';
                try {
                    sock.send(msg);
                }
                catch (e) {
                    error_handler(e)();
                }
            }
        },
        join: function(type, callback) {
            methods.send('join', type);
            room_callbacks[type] = room_callbacks[type] || [];
            room_callbacks[type].push(callback);
        },
        leave: function(type) {
            methods.send('leave', type);
            delete room_callbacks[type];
        },
        on_framerate: function () {},
        disconnect: terminate,
        ondisconnect: function (cb) {
            ondisconnect_callbacks.push(cb);
        },
        onconnect: function (cb) {
            onconnect_callbacks.push(cb);
        },
        poll: poll,
    };

    return methods;
}
